# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from pytest import approx
import numpy as np
import dask.array as da
from hyperspy.signals import Signal2D
import pyxem.data.dummy_data.dummy_data as dd
from pyxem.signals import BeamShift, LazyBeamShift, Diffraction2D
from hyperspy.axes import UniformDataAxis


class TestMakeLinearPlane:
    def test_simple(self):
        data_x, data_y = np.meshgrid(
            np.arange(-50, 50, dtype=np.float32), np.arange(-256, 0, dtype=np.float32)
        )
        data = np.stack((data_y, data_x), -1)
        s = BeamShift(data)
        s.change_dtype("float32")
        s_orig = s.deepcopy()
        s.make_linear_plane()
        assert s.data == approx(s_orig.data, abs=1e-6)


class TestGetLinearPlane:
    def test_simple(self):
        data_x, data_y = np.meshgrid(
            np.arange(-50, 50, dtype=np.float32), np.arange(-256, 0, dtype=np.float32)
        )
        data = np.stack((data_y, data_x), -1)
        s = BeamShift(data)
        s.change_dtype("float32")
        s_lp = s.get_linear_plane()
        assert s_lp.data == approx(s.data, abs=1e-6)

    def test_mask(self):
        data_x, data_y = np.meshgrid(
            np.arange(-50, 50, dtype=np.float32), np.arange(-256, 0, dtype=np.float32)
        )
        data = np.stack((data_y, data_x), -1)
        mask = np.zeros_like(data[:, :, 0], dtype=bool)
        mask[45:50, 36:41] = True
        s_mask = Signal2D(mask)
        s = BeamShift(data)
        s.change_dtype("float32")
        s_orig = s.deepcopy()
        s.data[45:50, 36:41, 0] = 100000
        s.data[45:50, 36:41, 1] = -100000
        s_lp = s.get_linear_plane(mask=s_mask)
        assert s_lp.data == approx(s_orig.data, abs=1e-6)

    def test_constrain_magnitude_variance(self):
        p = [0.5] * 6  # Plane parameters
        x, y = np.meshgrid(np.arange(256), np.arange(256))
        base_plane_x = p[0] * x + p[1] * y + p[2]
        base_plane_y = p[3] * x + p[4] * y + p[5]

        base_plane = np.stack((base_plane_x, base_plane_y)).T
        data = base_plane.copy()

        shifts = np.zeros_like(data)
        shifts[:128, 128:] = (10, 10)
        shifts[:128, :128] = (-10, -10)
        shifts[128:, 128:] = (-10, 10)
        shifts[128:, :128] = (10, 10)
        data += shifts

        s = BeamShift(data)

        s_lp = s.get_linear_plane()
        assert not np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

        s_lp = s.get_linear_plane(constrain_magnitude_variance=True)
        assert np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

    def test_constrain_magnitude_variance_mask(self):
        p = [0.5] * 6  # Plane parameters
        x, y = np.meshgrid(np.arange(256), np.arange(256))
        base_plane_x = p[0] * x + p[1] * y + p[2]
        base_plane_y = p[3] * x + p[4] * y + p[5]

        base_plane = np.stack((base_plane_x, base_plane_y)).T
        data = base_plane.copy()

        shifts = np.zeros_like(data)
        shifts[:128, 128:] = (10, 10)
        shifts[:128, :128] = (-10, -10)
        shifts[128:, 128:] = (-10, 10)
        shifts[128:, :128] = (10, 10)
        shifts[:, -10:] = (9999, 1321)
        shifts[:, :10] = (2213, -9879)
        data += shifts

        mask = np.zeros((256, 256), dtype=bool)
        mask[:, -10:] = True
        mask[:, :10] = True

        s = BeamShift(data)

        s_lp = s.get_linear_plane(constrain_magnitude_variance=True)
        assert not np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

        s_lp = s.get_linear_plane(constrain_magnitude_variance=True, mask=mask)
        assert np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

        # Check wrong mask dimensions
        mask = mask[:255]
        with pytest.raises(ValueError):
            s_lp = s.get_linear_plane(constrain_magnitude_variance=True, mask=mask)

    def test_constrain_magnitude_variance_initial_values(self):
        p = [0.5] * 6  # Plane parameters
        x, y = np.meshgrid(np.arange(256), np.arange(256))
        base_plane_x = p[0] * x + p[1] * y + p[2]
        base_plane_y = p[3] * x + p[4] * y + p[5]

        base_plane = np.stack((base_plane_x, base_plane_y)).T
        data = base_plane.copy()

        shifts = np.zeros_like(data)
        shifts[:128, 128:] = (10, 10)
        shifts[:128, :128] = (-10, -10)
        shifts[128:, :] = (-10, 10)
        data += shifts

        s = BeamShift(data)

        # Plane fitting does poorly here, likely due to not enough different domains
        s_lp = s.get_linear_plane(constrain_magnitude_variance=True)
        assert not np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

        # Varying the initial values around can help find different planes
        initial_values = [1.0] * 6
        s_lp = s.get_linear_plane(
            constrain_magnitude_variance=True, initial_values=initial_values
        )
        assert np.allclose(s_lp.data, base_plane.data, rtol=1e-7)

    def test_lazy_input_error(self):
        s = LazyBeamShift(da.zeros((50, 40, 2)))
        with pytest.raises(ValueError):
            s.get_linear_plane()
        s.compute()
        s.get_linear_plane()

    @pytest.mark.parametrize("constrain_magnitude_variance", [False, True])
    def test_wrong_navigation_input_1d_error(self, constrain_magnitude_variance):
        s = BeamShift(np.zeros((50, 2)))
        with pytest.raises(NotImplementedError):
            s.get_linear_plane(
                constrain_magnitude_variance=constrain_magnitude_variance
            )

    @pytest.mark.parametrize("constrain_magnitude_variance", [False, True])
    def test_wrong_navigation_input_3d_error(self, constrain_magnitude_variance):
        s = BeamShift(np.zeros((50, 40, 30, 2)))
        with pytest.raises(NotImplementedError):
            s.get_linear_plane(
                constrain_magnitude_variance=constrain_magnitude_variance
            )
        s1 = s.inav[:, :, 10]
        s1.get_linear_plane(constrain_magnitude_variance=constrain_magnitude_variance)

    def test_wrong_input_fit_corners_and_mask(self):
        s = BeamShift(np.zeros((5, 5, 2)))
        with pytest.raises(ValueError):
            s.get_linear_plane(fit_corners=0.05, mask=np.ones((5, 5)))

    def test_non_bool_mask(self):
        s = BeamShift(np.zeros((50, 40, 2)))
        mask = np.zeros((50, 40), dtype=np.int32)
        s_mask = Signal2D(mask)
        with pytest.raises(ValueError):
            s.get_linear_plane(mask=s_mask)
        s_mask.change_dtype(bool)
        s.get_linear_plane(mask=s_mask)


class TestBeamShiftFitCorners:
    def test_fit_corners_flat(self):
        data0 = np.ones(shape=(64, 64, 2))
        s0 = BeamShift(data0)
        s0_linear_plane = s0.get_linear_plane(fit_corners=0.05)
        s0_corr = s0 - s0_linear_plane
        assert (s0.data == data0).all()
        assert s0_corr.data == approx(0.0, abs=1e-7)

    def test_fit_corners_x_y(self):
        array_x, array_y = np.meshgrid(range(64), range(64))
        data_x = np.dstack((array_x, array_x)).astype("float64")
        data_y = np.dstack((array_y, array_y)).astype("float64")
        s_x = BeamShift(data_x)
        s_y = BeamShift(data_y)

        s_x_corr = s_x - s_x.get_linear_plane(fit_corners=0.05)
        s_y_corr = s_y - s_y.get_linear_plane(fit_corners=0.05)
        assert s_x_corr.data == approx(0.0, abs=1e-6)
        assert s_y_corr.data == approx(0.0, abs=1e-6)

        data_xy = np.dstack((array_x, array_y)).astype("float64")
        data_yx = np.dstack((array_y, array_x)).astype("float64")
        s_xy = BeamShift(data_xy)
        s_yx = BeamShift(data_yx)
        s_xy_corr = s_xy - s_xy.get_linear_plane(fit_corners=0.05)
        s_yx_corr = s_yx - s_yx.get_linear_plane(fit_corners=0.05)
        assert s_xy_corr.data == approx(0.0, abs=1e-6)
        assert s_yx_corr.data == approx(0.0, abs=1e-6)

        data_tilt = np.dstack((array_x + array_y, np.fliplr(array_x) + array_y))
        s_tilt = BeamShift(data_tilt.astype("float64"))
        s_tilt_corr = s_tilt - s_tilt.get_linear_plane(fit_corners=0.05)
        assert s_tilt_corr.data == approx(0.0, abs=1e-6)

    def test_fit_corners_one_large_value(self):
        array_x, array_y = np.meshgrid(range(64), range(64))
        array_x, array_y = array_x.astype("float64"), array_y.astype("float64")
        data = np.dstack((array_x + array_y, np.fliplr(array_x) + array_y))
        add_number_slice = np.s_[20:30, 30:40, :]
        data[add_number_slice] += 1000
        s = BeamShift(data)
        s_corr = s - s.get_linear_plane(fit_corners=0.05)
        s_corr.data[add_number_slice] -= 1000
        assert s_corr.data == approx(0.0, abs=1e-8)

    def test_calibrate_electric_shifts(self):
        data = np.zeros((100, 100, 2))
        data[45:55, :] = 3
        s = BeamShift(data)
        s.metadata.add_node("Shifts")  # add information about the signal Axes
        s.metadata.Shifts.signal_axes = (
            UniformDataAxis(units="mrad", scale=0.01, offset=0.0, name="x"),
            UniformDataAxis(units="mrad", scale=0.01, offset=0.0, name="y"),
        )

        cal = s.pixels_to_calibrated_units()
        cal.beam_energy = 200
        cal = cal.calibrate_electric_shifts(thickness=60)
        assert cal.data[0, :, 0] == approx(0)
        assert cal.data[45, :, 1] == approx(-1.7187, abs=1e-3)

    def test_calibrate_electric_shifts_failure_units(self):
        data = np.zeros((100, 100, 2))
        data[45:55, :] = 3
        s = BeamShift(data)
        s.metadata.add_node("Shifts")

        s.metadata.Shifts.signal_axes = (
            UniformDataAxis(units="nm^-1", scale=0.01, offset=0.0, name="x"),
            UniformDataAxis(units="nm^-1", scale=0.01, offset=0.0, name="y"),
        )

        cal = s.pixels_to_calibrated_units()
        with pytest.raises(ValueError):
            # This should raise an error because the beam energy is not set
            cal.calibrate_electric_shifts(thickness=60)

    def test_calibrate_electric_shifts_failure(self):
        data = np.zeros((100, 100, 2))
        data[45:55, :] = 3
        s = BeamShift(data)
        s.metadata.add_node("Shifts")

        s.metadata.Shifts.signal_axes = (
            UniformDataAxis(units="mrad", scale=0.01, offset=0.0, name="x"),
            UniformDataAxis(units="mrad", scale=0.01, offset=0.0, name="y"),
        )

        cal = s.pixels_to_calibrated_units()
        with pytest.raises(ValueError):
            # This should raise an error because the beam energy is not set
            cal.calibrate_electric_shifts(thickness=60)

    def test_large_values_not_in_corners(self):
        s = BeamShift(np.zeros((100, 100, 2)))
        s.isig[0].data[:5, :5], s.isig[0].data[:5, -5:] = 10, 10
        s.isig[0].data[-5:, :5], s.isig[0].data[-5:, -5:] = 10, 10
        s.isig[1].data[:5, :5], s.isig[1].data[:5, -5:] = 30, 30
        s.isig[1].data[-5:, :5], s.isig[1].data[-5:, -5:] = 30, 30

        cross_array = np.zeros((100, 100, 2))
        cross_array[30:70, :, :], cross_array[:, 30:70, :] = 1000, 1000
        s.data = s.data + cross_array

        s1 = s - s.get_linear_plane(fit_corners=0.05)
        s1.data = s1.data - cross_array

        assert s1.isig[0].data[5:95, :] == approx(np.ones((90, 100)) * -10)
        assert s1.isig[0].data[:, 5:95] == approx(np.ones((100, 90)) * -10)
        assert s1.isig[0].data[5:95, :] == approx(np.ones((90, 100)) * -10)
        assert s1.isig[1].data[5:95, :] == approx(np.ones((90, 100)) * -30)
        assert s1.isig[1].data[:, 5:95] == approx(np.ones((100, 90)) * -30)
        assert s1.isig[1].data[5:95, :] == approx(np.ones((90, 100)) * -30)
        assert s1.inav[:5, :5].data == approx(np.zeros((5, 5, 2)), abs=1e-7)
        assert s1.inav[-5:, :5].data == approx(np.zeros((5, 5, 2)), abs=1e-7)
        assert s1.inav[:5, -5:].data == approx(np.zeros((5, 5, 2)), abs=1e-7)
        assert s1.inav[-5:, -5:].data == approx(np.zeros((5, 5, 2)), abs=1e-7)

    def test_cropped_beam_shift_signal(self):
        s = BeamShift(np.random.random((200, 200, 2)))
        s_crop = s.inav[50:150, 50:150]
        s_crop_corr = s_crop - s_crop.get_linear_plane(fit_corners=0.05)
        assert s_crop_corr.axes_manager.shape == s_crop.axes_manager.shape

    def test_different_fit_corners_value(self):
        data = np.zeros((100, 100, 2))
        data[10:-10, 10:-10] = 100000
        s = BeamShift(data)
        s_lp_05 = s.get_linear_plane(fit_corners=0.05)
        s_lp_15 = s.get_linear_plane(fit_corners=0.15)
        assert s_lp_05.data == approx(0.0, abs=1e-8)
        assert s_lp_15.data != approx(0.0, abs=1e-8)


class TestFullDirectBeamCentering:
    def setup_method(self):
        data = np.zeros((3, 3, 16, 16), dtype=np.float32)
        data[0, 0, 6, 7] = 10
        data[0, 1, 6, 8] = 10
        data[0, 2, 6, 9] = 10
        data[1, 0, 7, 7] = 10
        data[1, 1, 7, 8] = 10
        data[1, 2, 7, 9] = 10
        data[2, 0, 8, 7] = 10
        data[2, 1, 8, 8] = 10
        data[2, 2, 8, 9] = 10
        s = Diffraction2D(data)
        s.axes_manager[0].scale = 0.5
        s.axes_manager[1].scale = 1.5
        s.axes_manager[2].scale = 5
        s.axes_manager[3].scale = 5
        s.axes_manager[0].offset = -10
        s.axes_manager[1].offset = 20
        s.axes_manager[2].offset = 30
        s.axes_manager[3].offset = 30
        self.s = s

    def test_simple(self):
        s = self.s
        s_beam_shift = s.get_direct_beam_position(method="blur", sigma=1)
        s.center_direct_beam(shifts=s_beam_shift)
        assert (s.data[:, :, 8, 8] == 10).all()
        s.data[:, :, 8, 8] = 0
        assert not s.data.any()

    def test_simple_1d(self):
        s = self.s.inav[0]
        s_beam_shift = s.get_direct_beam_position(method="blur", sigma=1)
        s.center_direct_beam(shifts=s_beam_shift)
        assert (s.data[:, 8, 8] == 10).all()
        s.data[:, 8, 8] = 0
        assert not s.data.any()

    def test_simple_lazy(self):
        s = self.s
        s = s.as_lazy()
        s_beam_shift = s.get_direct_beam_position(method="blur", sigma=1)
        s.center_direct_beam(shifts=s_beam_shift)
        s.compute()
        assert (s.data[:, :, 8, 8] == 10).all()
        s.data[:, :, 8, 8] = 0
        assert not s.data.any()

    def test_mask(self):
        s = self.s
        s.data[1, 2, 2, 3] = 1000
        mask = np.zeros((3, 3), dtype=bool)
        mask[1, 2] = True
        s_mask = Signal2D(mask)
        s_beam_shift = s.get_direct_beam_position(method="blur", sigma=1)
        s_linear_plane = s_beam_shift.get_linear_plane(mask=s_mask)
        s.center_direct_beam(shifts=s_linear_plane)
        assert s.data[:, :, 8, 8] == approx(10, abs=1e-5)
        s.data[:, :, 8, 8] = 0
        s.data[1, 2, 3, 2] = 0
        assert s.data == approx(0.0, abs=1e-4)

    def test_mask_lazy(self):
        s = self.s
        s.data[1, 2, 2, 3] = 1000
        s = self.s.as_lazy()
        mask = np.zeros((3, 3), dtype=bool)
        mask[1, 2] = True
        s_mask = Signal2D(mask)
        s_beam_shift = s.get_direct_beam_position(method="blur", sigma=1)
        s_beam_shift.compute()
        s_linear_plane = s_beam_shift.get_linear_plane(mask=s_mask)
        s.center_direct_beam(shifts=s_linear_plane)
        s.compute()
        assert s.data[:, :, 8, 8] == approx(10, abs=1e-5)
        s.data[:, :, 8, 8] = 0
        s.data[1, 2, 3, 2] = 0
        assert s.data == approx(0.0, abs=1e-4)


class TestGetMagnitudeSignal:
    def test_get_magnitude_signal_zeros(self):
        s = BeamShift(np.zeros((100, 100, 2)))
        s_magnitude = s.get_magnitude_signal()
        assert (s_magnitude.data == 0).all()

    def test_get_magnitude_signal_non_zero(self):
        data0 = np.zeros((100, 100, 2))
        data0[:, :, 0] = 1.5
        s0 = BeamShift(data0)
        s0_magnitude = s0.get_magnitude_signal()
        assert (s0_magnitude.data == 1.5).all()

        data1 = np.zeros((100, 100, 2))
        data1[:, :, 1] = 1.5
        s1 = BeamShift(data1)
        s1_magnitude = s1.get_magnitude_signal()
        assert (s1_magnitude.data == 1.5).all()

        data2 = np.zeros((100, 100, 2))
        data2[:, :, :] = 1.5
        s2 = BeamShift(data2)
        s2_magnitude = s2.get_magnitude_signal()
        assert s2_magnitude.data == approx(2.12132, abs=1e-5)

    def test_get_magnitude_single_value_non_zero(self):
        data = np.zeros((100, 50, 2))
        data[10, 15, 0] = 50
        s = BeamShift(data)
        s_magnitude = s.get_magnitude_signal(autolim=False)
        assert s_magnitude.data[10, 15] == 50
        s_magnitude.data[10, 15] = 0
        assert (s_magnitude.data == 0).all()

    def test_wrong_input_autolim_mangnitude_limits(self):
        s = BeamShift(np.zeros((10, 10, 2)))
        with pytest.raises(ValueError):
            s.get_magnitude_signal(autolim=True, magnitude_limits=(0, 30))


class TestGetPhaseSignal:
    def test_get_phase_signal(self):
        s = BeamShift(np.zeros((100, 100, 2)))
        s_p = s.get_phase_signal()
        assert s_p.axes_manager.shape == (100, 100)
        assert s_p.axes_manager.navigation_dimension == 0
        assert s_p.axes_manager.signal_dimension == 2

    def test_get_phase_signal_with_rotation(self):
        s = BeamShift(np.random.random(size=(100, 100, 2)))
        s.get_phase_signal(rotation=45)

    @pytest.mark.xfail(reason="Does not work with 1D navigation signals")
    def test_1d_get_phase_signal(self):
        s = BeamShift(np.zeros((100, 2)))
        s.get_phase_signal()


class TestGetColorSignal:
    def test_get_color_signal(self):
        s = BeamShift(np.random.random(size=(10, 10, 2)))
        s_color = s.get_color_signal()
        s_color.axes_manager.shape == (10, 10)


class TestGetMagnitudePhaseSignal:
    def test_get_magnitude_phase_signal(self):
        data_random = np.random.random(size=(64, 64, 2))
        s_random = BeamShift(data_random)
        s_random.get_magnitude_phase_signal()
        s_random.get_magnitude_phase_signal(rotation=45)
        s_random.get_magnitude_phase_signal(autolim=False, magnitude_limits=(0, 30))

    def test_get_magnitude_phase_signal_errors(self):
        data_random = np.random.random(size=(64, 64, 2))
        s_random = BeamShift(data_random)
        with pytest.raises(ValueError):
            s_random.get_magnitude_phase_signal(autolim=True, magnitude_limits=(0, 30))

    def test_get_magnitude_phase_signal_zeros(self):
        s = BeamShift(np.zeros((100, 100, 2)))
        s_color = s.get_magnitude_phase_signal()
        assert (s_color.data["R"] == 0).all()
        assert (s_color.data["G"] == 0).all()
        assert (s_color.data["B"] == 0).all()


class TestPlot:
    def test_simple_plot(self):
        s = BeamShift(np.zeros((10, 10, 2)))
        s.plot()

    def test_lazy_error(self):
        s = BeamShift(np.zeros((10, 10, 2))).as_lazy()
        with pytest.raises(ValueError):
            s.plot()


class TestGetBivariateHistogram:
    def test_get_bivariate_histogram_1d(self):
        s = BeamShift(np.random.random((10, 2)))
        s.get_bivariate_histogram()

    def test_get_bivariate_histogram_2d(self):
        s_random = BeamShift(np.random.random((64, 64, 2)))
        s_random.get_bivariate_histogram()

    def test_get_bivariate_histogram_3d(self):
        s_random = BeamShift(np.random.random((20, 64, 64, 2)))
        s_random.get_bivariate_histogram()

    def test_masked_get_bivariate_histogram(self):
        s = BeamShift(np.zeros((5, 5, 2)))
        value = 3
        s.data[0, 0, 0] = value
        s_hist = s.get_bivariate_histogram(bins=10, histogram_range=(-5, 5))
        assert s_hist.isig[3.0, 0.0].data[0] == 1.0
        assert s_hist.isig[0.0, 0.0].data[0] == (5 * 5) - 1

        masked = np.zeros((11, 11), dtype=bool)
        masked[0, 0] = True
        s_hist = s.get_bivariate_histogram(
            bins=10, histogram_range=(-5, 5), masked=masked
        )
        assert s_hist.isig[3.0, 0.0].data[0] == 0.0
        assert s_hist.isig[0.0, 0.0].data[0] == (5 * 5) - 1


class TestRotateBeamShifts:
    def setup_method(self):
        data = np.zeros((90, 70, 2))
        data[:, :, 0] = 1
        self.s = BeamShift(data)

    def test_clockwise_90_degrees(self):
        s = self.s
        s_rot = s.rotate_beam_shifts(90)
        data_x, data_y = s_rot.isig[0].data, s_rot.isig[1].data
        np.testing.assert_almost_equal(data_x, np.zeros_like(data_x))
        np.testing.assert_almost_equal(data_y, np.ones_like(data_y))

    def test_counterclockwise_90_degrees(self):
        s = self.s
        s_rot = s.rotate_beam_shifts(-90)
        data_x, data_y = s_rot.isig[0].data, s_rot.isig[1].data
        np.testing.assert_almost_equal(data_x, np.zeros_like(data_x))
        np.testing.assert_almost_equal(data_y, -np.ones_like(data_y))

    def test_180_degrees(self):
        s = self.s
        s_rot = s.rotate_beam_shifts(180)
        data_x, data_y = s_rot.isig[0].data, s_rot.isig[1].data
        np.testing.assert_almost_equal(data_x, -np.ones_like(data_x))
        np.testing.assert_almost_equal(data_y, np.zeros_like(data_y))

    def test_clockwise_45_degrees(self):
        s = self.s
        sin_rad = np.sin(np.deg2rad(45))
        s_rot = s.rotate_beam_shifts(45)
        data_x, data_y = s_rot.isig[0].data, s_rot.isig[1].data
        np.testing.assert_almost_equal(data_x, np.ones_like(data_x) * sin_rad)
        np.testing.assert_almost_equal(data_y, np.ones_like(data_y) * sin_rad)

    def test_counterclockwise_45_degrees(self):
        s = self.s
        sin_rad = np.sin(np.deg2rad(45))
        s_rot = s.rotate_beam_shifts(-45)
        data_x, data_y = s_rot.isig[0].data, s_rot.isig[1].data
        np.testing.assert_almost_equal(data_x, np.ones_like(data_x) * sin_rad)
        np.testing.assert_almost_equal(data_y, -np.ones_like(data_y) * sin_rad)


class TestRotateScanDimensions:
    def test_clockwise(self):
        s = dd.get_simple_beam_shift_signal()
        s_rot = s.rotate_scan_dimensions(1)
        assert not (s_rot.data[0, 0:10, 0] == 0).all()
        assert (s_rot.data[0, -10:, 0] == 0).all()
        assert not (s_rot.data[-10:, 0, 0] == 0).all()
        assert (s_rot.data[0:10, 0, 0] == 0).all()

    def test_counterclockwise(self):
        s = dd.get_simple_beam_shift_signal()
        s_rot = s.rotate_scan_dimensions(-1)
        assert (s_rot.data[0, 0:10, 0] == 0).all()
        assert not (s_rot.data[0, -10:, 0] == 0).all()
        assert (s_rot.data[-10:, 0, 0] == 0).all()
        assert not (s_rot.data[0:10, 0, 0] == 0).all()


class TestPhaseRetrieval:
    def setup_method(self):
        # construct the surface, two point with Gaussian distribution
        coords = np.linspace(-20, 10, num=512)
        x, y = np.meshgrid(coords, coords)
        surface = np.exp(-(x**2 + y**2) / 2) + np.exp(
            -((x - 2) ** 2 + (y + 4) ** 2) / 2
        )

        # x and y phase gradient of the Gaussians, analytical form
        dx = x * (-np.exp(-(x**2) / 2 - y**2 / 2)) + (x - 2) * -np.exp(
            (-0.5 * (x - 2) ** 2 - 0.5 * (y + 4) ** 2)
        )
        dy = y * (-np.exp(-(x**2) / 2 - y**2 / 2)) + (y + 4) * -np.exp(
            (-0.5 * (x - 2) ** 2 - 0.5 * (y + 4) ** 2)
        )

        data = np.empty((512, 512, 2))
        data[:, :, 0] = dx
        data[:, :, 1] = dy
        s = BeamShift(data)
        s.axes_manager.navigation_axes[0].axis = coords
        s.axes_manager.navigation_axes[1].axis = coords
        self.s = s

        # normalise for comparison later
        surface -= surface.mean()
        surface /= surface.std()
        self.surface = surface

    @pytest.mark.parametrize("method", ["kottler", "arnison", "frankot"])
    def test_kottler(self, method):
        s_recon = self.s.phase_retrieval(method=method)
        recon = s_recon.data
        recon -= recon.mean()
        recon /= recon.std()

        assert np.isclose(self.surface, recon, atol=1e-3).all()

    @pytest.mark.parametrize("method", ["kottler", "arnison", "frankot"])
    def test_mirroring(self, method):
        s_recon = self.s.phase_retrieval(method, mirroring=True)
        recon = s_recon.data
        recon -= recon.mean()
        recon /= recon.std()

        assert np.isclose(self.surface, recon, atol=1e-3).all()

    @pytest.mark.parametrize("method", ["kottler", "arnison", "frankot"])
    def test_mirror_flip(self, method):
        s_noflip = self.s.phase_retrieval(method, mirroring=True, mirror_flip=False)
        s_flip = self.s.phase_retrieval(method, mirroring=True, mirror_flip=True)
        noflip = s_noflip.data
        noflip -= noflip.mean()
        noflip /= noflip.std()
        flip = s_flip.data
        flip -= flip.mean()
        flip /= flip.std()

        noflip_sum_diff = np.abs(self.surface - noflip).sum()
        flip_sum_diff = np.abs(self.surface - flip).sum()

        assert noflip_sum_diff != flip_sum_diff

    @pytest.mark.xfail(reason="invalid_method")
    def test_unavailable_method(self):
        self.s.phase_retrieval("magic!")



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from pyxem.utils import indexation_utils as iutls
from pyxem.utils import polar_transform_utils as ptutls
import numpy as np
import pytest
from unittest.mock import Mock
import dask.array as da
import sys
from pyxem.utils.cuda_utils import is_cupy_array


try:
    import cupy as cp

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
    cp = np

skip_cupy = pytest.mark.skipif(not CUPY_INSTALLED, reason="cupy is required")


@pytest.fixture()
def simulations():
    mock_sim_1 = Mock()
    mock_sim_1.calibrated_coordinates = np.array(
        [
            [3, 4, 0],  # 5
            [5, 12, 0],  # 13
            [8, 15, 0],  # 17
            [-8, 15, 0],  # 17
        ]
    )
    mock_sim_1.intensities = np.array([2, 3, 4, 2])
    mock_sim_2 = Mock()
    mock_sim_2.calibrated_coordinates = np.array(
        [[5, 12, 0], [8, 15, 0], [7, 24, 0]]  # 13  # 17  # 25
    )
    mock_sim_2.intensities = np.array([1, 2, 10])
    mock_sim_3 = Mock()
    mock_sim_3.calibrated_coordinates = np.array(
        [
            [3, 4, 0],  # 5
            [5, 12, 0],  # 13
        ]
    )
    mock_sim_3.intensities = np.array([8, 1])
    simlist = [mock_sim_1, mock_sim_2, mock_sim_3]
    return simlist


@pytest.fixture()
def mock_sim():
    xx = np.linspace(-10, 11, 5)
    yy = np.linspace(-10, 11, 3)
    XX, YY = np.meshgrid(xx, yy)
    Z = np.zeros(XX.ravel().shape)
    mock_sim = Mock()
    mock_sim.calibrated_coordinates = np.vstack([XX.ravel(), YY.ravel(), Z]).T
    mock_sim.intensities = np.ones(Z.shape)
    return mock_sim


@pytest.mark.parametrize(
    "rot, dr, dt, nim, nt",
    [
        (20, 1, 1, True, True),
        (50, 0.1, 1, False, True),
        (180, 0.1, 0.1, True, False),
        (270, 0.1, 0.1, False, False),
    ],
)
def test_match_image_to_template(mock_sim, rot, dr, dt, nim, nt):
    # serves to actually validate the correctness
    x, y, i = ptutls.get_template_cartesian_coordinates(
        mock_sim, in_plane_angle=rot, center=(20, 20)
    )
    # generate an image from the template
    tim = np.zeros((40, 40))
    y = 40 - y
    tim[np.rint(y).astype(np.int32), np.rint(x).astype(np.int32)] = 1.0
    # compare image and template and retrieve the same angle
    a, c = iutls.get_in_plane_rotation_correlation(
        tim,
        mock_sim,
        find_direct_beam=False,
        delta_r=dr,
        delta_theta=dt,
        normalize_image=nim,
        normalize_template=nt,
    )
    assert abs(a[np.argmax(c)] - rot) % 180 < 2.0


@pytest.mark.parametrize(
    "rot, dr, dt, nim, nt",
    [
        (20, 1, 1, True, True),
        (50, 0.1, 1, False, True),
        (180, 0.1, 0.1, True, False),
        (270, 0.1, 0.1, False, False),
    ],
)
@skip_cupy
def test_match_image_to_template_gpu(mock_sim, rot, dr, dt, nim, nt):
    # serves to actually validate the correctness
    x, y, i = ptutls.get_template_cartesian_coordinates(
        mock_sim, in_plane_angle=rot, center=(20, 20)
    )
    # generate an image from the template
    tim = cp.zeros((40, 40))
    y = 40 - y
    tim[np.rint(y).astype(np.int32), np.rint(x).astype(np.int32)] = 1.0
    # compare image and template and retrieve the same angle
    a, c = iutls.get_in_plane_rotation_correlation(
        tim,
        mock_sim,
        find_direct_beam=False,
        delta_r=dr,
        delta_theta=dt,
        normalize_image=nim,
        normalize_template=nt,
    )
    assert abs(a[cp.argmax(c)] - rot) % 180 < 2.0


@pytest.mark.parametrize(
    "max_radius, expected_shapes, test_pos, test_int",
    [
        (None, ((3, 2, 4), (3, 4)), np.array((-8, 15)), 2),
        (15, ((3, 2, 4), (3, 4)), np.array((0, 0)), 0),
    ],
)
def test_simulations_to_arrays(
    simulations, max_radius, expected_shapes, test_pos, test_int
):
    positions, intensities = iutls._simulations_to_arrays(
        simulations, max_radius=max_radius
    )
    np.testing.assert_array_equal(positions.shape, expected_shapes[0])
    np.testing.assert_array_equal(intensities.shape, expected_shapes[1])
    np.testing.assert_array_equal(positions[0, :, -1], test_pos)
    np.testing.assert_array_equal(intensities[0, -1], test_int)


def test_match_polar_to_polar_template():
    image = np.ones((123, 50))
    r = np.linspace(2, 40, 30, dtype=np.int32)
    theta = np.linspace(10, 110, 30, dtype=np.int32)
    intensities = np.ones(30, dtype=np.float64)
    cor = iutls._match_polar_to_polar_template(image, r, theta, intensities)
    np.testing.assert_array_almost_equal(cor, np.ones(123, dtype=np.float64) * 30)


@skip_cupy
def test_match_polar_to_polar_template_gpu():
    image = cp.ones((123, 50))
    r = cp.linspace(2, 40, 30, dtype=np.int32)
    theta = cp.linspace(10, 110, 30, dtype=np.int32)
    intensities = cp.ones(30, dtype=np.float64)
    cor = iutls._match_polar_to_polar_template(image, r, theta, intensities)
    cp.testing.assert_array_almost_equal(cor, cp.ones(123, dtype=np.float64) * 30)


@pytest.mark.parametrize(
    "itf, norim, nortemp",
    [
        (None, True, False),
        (np.sqrt, False, True),
    ],
)
def test_get_in_plane_rotation_correlation(simulations, itf, norim, nortemp):
    image = np.ones((123, 50))
    simulation = simulations[0]
    ang, cor = iutls.get_in_plane_rotation_correlation(
        image,
        simulation,
        intensity_transform_function=itf,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_template=nortemp,
    )
    assert ang.shape[0] == 277
    assert cor.shape[0] == 277


@skip_cupy
@pytest.mark.parametrize(
    "itf, norim, nortemp",
    [
        (None, True, False),
        (cp.sqrt, False, True),
    ],
)
def test_get_in_plane_rotation_correlation_gpu(simulations, itf, norim, nortemp):
    image = cp.ones((123, 50))
    simulation = simulations[0]
    ang, cor = iutls.get_in_plane_rotation_correlation(
        image,
        simulation,
        intensity_transform_function=itf,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_template=nortemp,
    )
    assert ang.shape[0] == 277
    assert cor.shape[0] == 277


@pytest.mark.parametrize(
    "N, n_keep, frac_keep, res",
    [
        (100, None, None, 100),
        (100, None, 0.392, 39),
        (100, 20, None, 20),
        (100, 20, 0.5, 20),
    ],
)
def test_get_max_n(N, n_keep, frac_keep, res):
    np.testing.assert_almost_equal(res, iutls._get_max_n(N, n_keep, frac_keep))


@pytest.mark.parametrize(
    "dtype",
    [
        (np.float64),
        (np.float32),
    ],
)
def test_match_polar_to_polar_library(dtype):
    image = np.ones((123, 50), dtype=dtype)
    r = np.linspace(2, 40, 30, dtype=np.int32)
    r = np.repeat(r[np.newaxis, ...], 4, axis=0)
    theta = np.linspace(10, 110, 30, dtype=np.int32)
    theta = np.repeat(theta[np.newaxis, ...], 4, axis=0)
    intensities = np.ones(30, dtype=dtype)
    intensities = np.repeat(intensities[np.newaxis, ...], 4, axis=0)
    angles, cor, angles_m, cor_m = iutls._get_full_correlations(
        image,
        r,
        theta,
        intensities,
    )
    assert cor.shape[0] == 4
    assert angles.shape[0] == 4
    assert cor_m.shape[0] == 4
    assert angles_m.shape[0] == 4
    assert cor.dtype == dtype
    assert angles.dtype == np.int32
    assert cor_m.dtype == dtype
    assert angles_m.dtype == np.int32


@pytest.mark.parametrize(
    "norim, nortemp",
    [
        (True, False),
        (False, True),
    ],
)
def test_correlate_library_to_pattern(simulations, norim, nortemp):
    image = np.ones((123, 50))
    index, ang, cor, ang_m, cor_m = iutls.correlate_library_to_pattern(
        image,
        simulations,
        delta_r=2.6,
        delta_theta=1.3,
        frac_keep=1.0,
        normalize_image=norim,
        normalize_templates=nortemp,
    )
    assert ang.shape[0] == 3
    assert cor.shape[0] == 3
    assert ang_m.shape[0] == 3
    assert cor_m.shape[0] == 3
    assert index.dtype == np.int32
    assert ang.dtype == image.dtype
    assert cor.dtype == image.dtype
    assert ang_m.dtype == image.dtype
    assert cor_m.dtype == image.dtype


@pytest.mark.parametrize(
    "normed",
    [True, False],
)
def test_get_integrated_templates(normed):
    rmax = 50
    r = np.linspace(2, 40, 30, dtype=np.uint64)
    r = np.repeat(r[np.newaxis, ...], 4, axis=0)
    intensities = np.ones(30, dtype=np.float64)
    intensities = np.repeat(intensities[np.newaxis, ...], 4, axis=0)
    integrated = iutls._get_integrated_polar_templates(rmax, r, intensities, normed)
    assert integrated.shape[0] == 4
    assert integrated.shape[1] == 50


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
def test_match_library_to_polar_fast():
    polar_sum = np.ones(50, dtype=np.float64)
    integrated_temp = np.arange(0, 50, dtype=np.float64)
    integrated_temp = np.repeat(integrated_temp[np.newaxis, ...], 5, axis=0)
    coors = iutls._match_library_to_polar_fast(
        polar_sum,
        integrated_temp,
    )
    assert coors.shape[0] == 5
    assert coors.ndim == 1


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.parametrize(
    "norim, nortemp",
    [
        (True, False),
        (False, True),
    ],
)
def test_correlate_library_to_pattern_fast(simulations, norim, nortemp):
    image = np.ones((123, 50))
    cor = iutls.correlate_library_to_pattern_fast(
        image,
        simulations,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_templates=nortemp,
    )
    assert cor.shape[0] == 3


@pytest.mark.parametrize(
    "norim, nortemp",
    [
        (True, False),
        (False, True),
    ],
)
@skip_cupy
def test_correlate_library_to_pattern_fast_gpu(simulations, norim, nortemp):
    image = cp.ones((123, 50))
    cor = iutls.correlate_library_to_pattern_fast(
        image,
        simulations,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_templates=nortemp,
    )
    assert cor.shape[0] == 3
    assert is_cupy_array(cor)


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.parametrize("intensity_transform_function", [np.sqrt, None])
def test_prep_image_and_templates(simulations, intensity_transform_function):
    image = np.ones((123, 51))
    pim, r, t, i = iutls._prepare_image_and_templates(
        image,
        simulations,
        2.1,
        3.2,
        33.3,
        intensity_transform_function,
        True,
        None,
        True,
        True,
    )
    assert pim.shape[0] == 112
    assert pim.shape[1] == 16
    assert r.max() <= 15
    assert t.min() >= 0
    assert t.max() < 112
    assert i.shape[0] == 3
    assert i.shape[1] == 4


@skip_cupy
@pytest.mark.parametrize("intensity_transform_function", [cp.sqrt, None])
def test_prep_image_and_templates_gpu(simulations, intensity_transform_function):
    image = cp.ones((123, 51))
    pim, r, t, i = iutls._prepare_image_and_templates(
        image,
        simulations,
        2.1,
        3.2,
        33.3,
        intensity_transform_function,
        True,
        None,
        True,
        True,
    )
    assert pim.shape[0] == 112
    assert pim.shape[1] == 16
    assert r.max() <= 15
    assert t.min() >= 0
    assert t.max() < 112
    assert i.shape[0] == 3
    assert i.shape[1] == 4


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.parametrize(
    "nbest",
    [1, 2],
)
@pytest.mark.slow
def test_mixed_matching_lib_to_polar(nbest):
    image = np.ones((123, 51), dtype=np.float64)
    polar_norm = 1.0
    polar_sum = np.sum(image, axis=0)
    polar_sum_norm = 2.3
    integrated_templates = np.ones((5, 51), dtype=np.float64)
    integrated_template_norms = np.sum(integrated_templates, axis=1)
    r = np.linspace(2, 40, 30, dtype=np.int64)
    r = np.repeat(r[np.newaxis, ...], 5, axis=0)
    theta = np.linspace(10, 110, 30, dtype=np.int64)
    theta = np.repeat(theta[np.newaxis, ...], 5, axis=0)
    intensities = np.ones(30, dtype=np.float64)
    intensities = np.repeat(intensities[np.newaxis, ...], 5, axis=0)
    template_norms = np.ones(5, dtype=np.float64) * 3
    fraction = 0.6
    answer = iutls._mixed_matching_lib_to_polar(
        image,
        integrated_templates,
        r,
        theta,
        intensities,
        None,
        1.0,
        nbest,
    )
    assert answer.shape[0] == nbest
    assert answer.shape[1] == 4


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.parametrize(
    "nbest, frk, norim, nortemp",
    [
        (3, None, True, False),
        (1, 0.5, False, True),
    ],
)
@pytest.mark.slow
def test_get_n_best_matches(simulations, nbest, frk, norim, nortemp):
    image = np.ones((123, 50))
    indx, angs, cor, signs = iutls.get_n_best_matches(
        image,
        simulations,
        n_best=nbest,
        frac_keep=frk,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_templates=nortemp,
    )
    assert cor.shape[0] == indx.shape[0] == angs.shape[0] == nbest


@skip_cupy
@pytest.mark.parametrize(
    "nbest, frk, norim, nortemp",
    [
        (3, None, True, False),
        (1, 0.5, False, True),
    ],
)
@pytest.mark.slow
def test_get_n_best_matches_gpu(simulations, nbest, frk, norim, nortemp):
    image = cp.ones((123, 50))
    indx, angs, cor, signs = iutls.get_n_best_matches(
        image,
        simulations,
        n_best=nbest,
        frac_keep=frk,
        delta_r=2.6,
        delta_theta=1.3,
        normalize_image=norim,
        normalize_templates=nortemp,
    )
    assert cor.shape[0] == indx.shape[0] == angs.shape[0] == nbest


def create_dataset(shape):
    dataset = Mock()
    data = np.ones(shape)
    dataset.data = da.from_array(data)
    dataset._lazy = True
    dataset.axes_manager.navigation_dimension = len(shape) - 2
    dataset.axes_manager.signal_indices_in_array = (len(shape) - 2, len(shape) - 1)
    return dataset


@pytest.fixture()
def library():
    library = {}
    mock_sim_1 = Mock()
    mock_sim_1.calibrated_coordinates = np.array(
        [
            [0, 1, 0],
            [1, 0, 0],
            [1, 2, 0],
            [-1, 2, 0],
        ]
    )
    mock_sim_1.intensities = np.array([2, 3, 4, 2])
    mock_sim_2 = Mock()
    mock_sim_2.calibrated_coordinates = np.array([[-1, -2, 0], [1, 2, 0], [-2, 1, 0]])
    mock_sim_2.intensities = np.array([1, 2, 10])
    simlist = [mock_sim_1, mock_sim_2]
    orientations = np.array(
        [
            [1, 2, 3],
            [3, 4, 5],
        ]
    )
    library["dummyphase"] = {"simulations": simlist, "orientations": orientations}
    return library


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.parametrize(
    "sigdim, n_best, frac_keep, norim, nort, chu",
    [
        ((2, 3), 1, 0.5, True, False, None),
        ((2, 3, 4), 1, 0.5, False, True, "auto"),
        ((2, 3, 4, 5), 2, 0.5, False, False, {0: "auto", 1: "auto", 2: None, 3: None}),
        ((2, 3, 4, 5), 9, 0.5, False, False, {0: "auto", 1: "auto", 2: None, 3: None}),
    ],
)
@pytest.mark.slow
def test_index_dataset_with_template_rotation(
    library, sigdim, n_best, frac_keep, norim, nort, chu
):
    signal = create_dataset(sigdim)
    result = iutls.index_dataset_with_template_rotation(
        signal,
        library,
        n_best=n_best,
        frac_keep=frac_keep,
        delta_r=0.5,
        delta_theta=36,
        max_r=None,
        intensity_transform_function=np.sqrt,
        normalize_images=norim,
        normalize_templates=nort,
        chunks=chu,
    )


# @pytest.mark.skipif(sys.platform=='darwin',reason="Fails on Mac OSX")
@pytest.mark.slow
def test_fail_index_dataset_with_template_rot(library):
    signal = create_dataset((3, 2, 4, 1, 2))
    with pytest.raises(ValueError):
        result = iutls.index_dataset_with_template_rotation(
            signal,
            library,
        )


@pytest.mark.parametrize(
    "sigdim, maxr, n_best, frac_keep, norim, nort, chu",
    [
        (
            (2, 3, 4, 5),
            None,
            1,
            1,
            False,
            False,
            {0: "auto", 1: "auto", 2: None, 3: None},
        ),
        ((2, 3, 4, 5), 3, 2, 1, False, False, {0: "auto", 1: "auto", 2: None, 3: None}),
        (
            (2, 3, 4, 5),
            9,
            3,
            0.5,
            False,
            False,
            {0: "auto", 1: "auto", 2: None, 3: None},
        ),
    ],
)
@pytest.mark.slow
@skip_cupy
def test_index_dataset_with_template_rotation_gpu(
    library, sigdim, maxr, n_best, frac_keep, norim, nort, chu
):
    signal = create_dataset(sigdim)
    result, dicto = iutls.index_dataset_with_template_rotation(
        signal,
        library,
        n_best=n_best,
        frac_keep=frac_keep,
        delta_r=0.5,
        delta_theta=36,
        max_r=maxr,
        intensity_transform_function=np.sqrt,
        normalize_images=norim,
        normalize_templates=nort,
        chunks=chu,
        target="gpu",
    )


def run_index_chunk(di, n_best, frac_keep):
    max_r = 4
    max_theta = 5
    di.random.seed(1001)
    data = di.random.random((2, 3, 10, 12))
    lib_n = 4
    lib_spots = 5
    r = di.random.randint(0, max_r, size=(lib_n, lib_spots))
    theta = di.random.randint(0, max_theta, size=(lib_n, lib_spots))
    intensities = di.random.random((lib_n, lib_spots))
    integrated_templates = di.random.random((lib_n, max_r))
    center = (7, 4)
    max_radius = max_r
    precision = np.float32
    answer = iutls._index_chunk(
        data,
        center,
        max_radius,
        (5, 4),
        precision,
        integrated_templates,
        r,
        theta,
        intensities,
        None,
        frac_keep,
        n_best,
        True,
    )
    assert answer.shape == (2, 3, n_best, 4)


@pytest.mark.parametrize(
    "n_best, fraction",
    [
        (1, 1),
        (2, 0),
        (1, 0.5),
    ],
)
def test_index_chunk(n_best, fraction):
    run_index_chunk(np, n_best, fraction)


@pytest.mark.parametrize(
    "n_best, fraction",
    [
        (1, 1),
        (2, 0),
        (1, 0.5),
    ],
)
@skip_cupy
def test_index_chunk_gpu(n_best, fraction):
    run_index_chunk(cp, n_best, fraction)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""PDF generator and associated tools."""

import numpy as np

from pyxem.signals import PairDistributionFunction1D
from pyxem.utils._signals import _transfer_navigation_axes
from pyxem.utils._deprecated import deprecated


class PDFGenerator1D:
    """Generates a PairDistributionFunction1D signal from a specified
    ReducedIntensity1D signal.

    Parameters
    ----------
    signal : ReducedIntensity1D
        A reduced intensity radial profile.
    """

    @deprecated(
        since="0.15",
        alternative="pyxem.signals.diffraction2d.get_pdf",
        removal="1.0.0",
    )
    def __init__(self, signal, *args, **kwargs):
        self.signal = signal

    @deprecated(
        since="0.15",
        alternative="pyxem.signals.diffraction2d.get_pdf",
        removal="1.0.0",
    )
    def get_pdf(self, s_min, s_max=None, r_min=0, r_max=20, r_increment=0.01):
        """Calculates the pdf from the reduced intensity signal.

        Parameters
        ----------
        s_min : float
            Minimum scattering vector s for the pdf calculation. Note that s is
            defined here as s = 2 sin(theta)/lambda = 1/d.
        s_max : float
            Maximum scattering vector s for the pdf calculation. Note that s is
            defined here as s = 2 sin(theta)/lambda = 1/d.
        r_cutoff : list of float
            A list with the format [<r_min>, <r_max>], which sets the
            limits of the real space axis in the calculated PDF.
        r_increment : float
            Step size in r in the extracted PDF.

        Returns
        -------
        pdf : PDF1D
            A signal of pair distribution functions.
        """
        s_scale = self.signal.axes_manager.signal_axes[0].scale
        if s_max is None:
            s_max = self.signal.axes_manager.signal_axes[0].size * s_scale
            print("s_max set to maximum of signal.")

        r_values = np.arange(r_min, r_max, r_increment)
        r_values = r_values.reshape(1, r_values.size)
        s_limits = [int(s_min / s_scale), int(s_max / s_scale)]

        # check that these aren't out of bounds
        if s_limits[1] > self.signal.axes_manager.signal_axes[0].size:
            raise ValueError(
                "User specified s_max is larger than the maximum "
                "scattering vector magnitude in the data. Please reduce "
                "s_max or use s_max=None to use the full scattering range."
            )
        s_values = np.arange(s_limits[0], s_limits[1], 1) * s_scale
        s_values = s_values.reshape(s_values.size, 1)  # column vector

        limited_red_int = self.signal.isig[s_limits[0] : s_limits[1]].data

        pdf_sine = np.sin(2 * np.pi * np.matmul(s_values, r_values))
        # creates a vector of the pdf
        rpdf = PairDistributionFunction1D(
            8 * np.pi * s_scale * np.matmul(limited_red_int, pdf_sine)
        )

        signal_axis = rpdf.axes_manager.signal_axes[0]
        signal_axis.scale = r_increment
        signal_axis.name = "Radius r"
        signal_axis.units = "$Å$"
        rpdf = _transfer_navigation_axes(rpdf, self.signal)

        title = self.signal.metadata.General.title
        rpdf.metadata.General.title = f"Pair distribution function of {title}"

        return rpdf



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from pyxem.utils._signals import _select_method_from_method_dict


@pytest.fixture()
def method_dict():
    return {"dummy_choice": _select_method_from_method_dict}


def test_select_fake_method_from_method_dict(method_dict):
    with pytest.raises(NotImplementedError):
        _ = _select_method_from_method_dict("fake_choice", method_dict, print_help=True)


def test__select_method_from_method_dict_print_help(method_dict):
    method_dict = {"dummy_choice": _select_method_from_method_dict}
    _ = _select_method_from_method_dict("dummy_choice", method_dict, print_help=True)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy.signals import Signal1D
from hyperspy._signals.lazy import LazySignal

import numpy as np
from fractions import Fraction as frac
from pyxem.utils._correlations import _get_interpolation_matrix, _symmetry_stem


class Correlation1D(Signal1D):
    """Signal class for pearson correlation and symmetry coefficient.

    Parameters
    ----------
    *args
        See :class:`~hyperspy._signals.signal1d.Signal1D`.
    **kwargs
        See :class:`~hyperspy._signals.signal1d.Signal1D`
    """

    _signal_type = "correlation"

    def get_symmetry_coefficient(
        self,
        angular_range=0.1,
        symmetries=[2, 3, 4, 5, 6, 7, 8, 9, 10],
        method="average",
        include_duplicates=False,
        normalize=True,
    ):
        """Return symmetry coefficient from pearson correlation function at all real
        space positions (n from 2 to 10).

        Parameters
        ----------
        angular_range: float
            The angular range (in rad) to integrate over when calculating the symmetry coefficient.
        symmetries: list
            The list of symmetries to calculate.
        method: "average", "first", "max"
            The method for calculating the Symmetry STEM
        include_duplicates: bool
            If angles which are duplicated should be included.
        normalize: bool
            This normalized by dividing by the number of angles for each symmetry.

        Returns
        -------
        sn: Signal1D
            Symmetry coefficient
        """
        angles = [set(frac(j, i) for j in range(1, i)) for i in symmetries]

        if not include_duplicates:  # remove duplicated symmetries
            already_used = set()
            new_angles = []
            for a in angles:
                new_angles.append(a.difference(already_used))
                already_used = already_used.union(a)
            angles = new_angles
        num_angles = [len(a) for a in angles]

        interp = [
            _get_interpolation_matrix(
                a,
                angular_range,
                num_points=self.axes_manager.signal_axes[0].size,
                method=method,
            )
            for a in angles
        ]
        signals = self.map(
            _symmetry_stem,
            interpolation=interp,
            show_progressbar=True,
            inplace=False,
            method=method,
            silence_warnings=True,
        )
        if method == "max" or method == "first":
            normalize = False
        if normalize:
            signals.data = np.divide(signals.data, num_angles)

        signals.axes_manager.signal_axes[-1].name = "Symmetry Order"
        signals.axes_manager.signal_axes[0].scale = 1
        signals.axes_manager.signal_axes[0].name = "Symmetry"
        signals.axes_manager.signal_axes[0].unit = "a.u."
        signals.axes_manager.signal_axes[0].offset = symmetries[0]

        return signals


class LazyCorrelation1D(LazySignal, Correlation1D):
    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains utility functions for processing electron diffraction
patterns.
"""

import numpy as np
import scipy.ndimage as ndi
import pyxem as pxm  # for ElectronDiffraction2D

from scipy.interpolate import interp1d
from scipy.signal import fftconvolve
from skimage import transform as tf
from skimage.feature import match_template
from skimage import morphology, filters
from skimage.draw import ellipse_perimeter
from skimage.registration import phase_cross_correlation
from tqdm import tqdm
from packaging.version import Version

from pyxem.utils.cuda_utils import is_cupy_array
from pyxem.utils._deprecated import deprecated
import pyxem.utils._pixelated_stem_tools as pst

try:
    import cupy as cp
    import cupyx.scipy.ndimage as ndigpu

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
    cp = None
    ndigpu = None

new_float_type = {
    # preserved types
    np.float32().dtype.char: np.float32,
    np.float64().dtype.char: np.float64,
    np.complex64().dtype.char: np.complex64,
    np.complex128().dtype.char: np.complex128,
    # altered types
    np.float16().dtype.char: np.float32,
    "g": np.float64,  # np.float128 ; doesn't exist on windows
    "G": np.complex128,  # np.complex256 ; doesn't exist on windows
}


def _index_coords(z, origin=None):
    """Creates x & y coords for the indices in a numpy array.

    Parameters
    ----------
    z : numpy.ndarray
        Two-dimensional data array containing signal.
    origin : tuple
        (x,y) defaults to the center of the image. Specify origin=(0,0) to set
        the origin to the *top-left* corner of the image.

    Returns
    -------
    x, y : arrays
        Corrdinates for the indices of a numpy array.
    """
    ny, nx = z.shape[:2]
    if origin is None:
        origin_x, origin_y = nx // 2, ny // 2
    else:
        origin_x, origin_y = origin

    x, y = np.meshgrid(np.arange(float(nx)), np.arange(float(ny)))

    x -= origin_x
    y -= origin_y
    return x, y


def _cart2polar(x, y):
    """Transform Cartesian coordinates to polar coordinates.

    Parameters
    ----------
    x, y : floats or arrays
        Cartesian coordinates

    Returns
    -------
    r, theta : floats or arrays
        Polar coordinates

    """
    r = np.sqrt(x**2 + y**2)
    theta = -np.arctan2(y, x)  # θ = 0 horizontal, +ve = anticlockwise
    return r, theta


def _polar2cart(r, theta):
    """Transform polar coordinates to Cartesian coordinates.

    Parameters
    ----------
    r, theta : floats or arrays
        Polar coordinates

    Returns
    -------
    x, y : floats or arrays
        Cartesian coordinates
    """
    # +ve quadrant in bottom right corner when plotted
    x = r * np.cos(theta)
    y = -r * np.sin(theta)
    return x, y


def gain_normalise(z, dref, bref):
    """Apply gain normalization to experimentally acquired electron
    diffraction pattern.

    Parameters
    ----------
    z : numpy.ndarray
        Two-dimensional data array containing signal.
    dref : ElectronDiffraction2D
        Two-dimensional data array containing dark reference.
    bref : ElectronDiffraction2D
        Two-dimensional data array containing bright reference.

    Returns
    -------
    z1 : np.array()
        Two dimensional data array of gain normalized z.
    """
    return ((z - dref) / (bref - dref)) * np.mean((bref - dref))


def remove_dead(z, deadpixels):
    """Remove dead pixels from experimental electron diffraction patterns.

    Parameters
    ----------
    z : np.array()
        Two-dimensional data array containing signal.
    deadpixels : np.array()
        Array containing the array indices of dead pixels in the diffraction
        pattern.

    Returns
    -------
    img : array
        Two-dimensional data array containing z with dead pixels removed.
    """
    z_bar = np.copy(z)
    for i, j in deadpixels:
        z_bar[i, j] = (z[i - 1, j] + z[i + 1, j] + z[i, j - 1] + z[i, j + 1]) / 4

    return z_bar


def convert_affine_to_transform(D, shape):
    """Converts an affine transform on a diffraction pattern to a suitable
    form for :func:`skimage.transform.warp`

    Parameters
    ----------
    D : numpy.ndarray
        Affine transform to be applied
    shape : tuple
        Shape tuple in form (y,x) for the diffraction pattern

    Returns
    -------
    transformation : numpy.ndarray
        3x3 numpy array of the transformation to be applied.

    """

    shift_x = (shape[1] - 1) / 2
    shift_y = (shape[0] - 1) / 2

    tf_shift = tf.SimilarityTransform(translation=[-shift_x, -shift_y])
    tf_shift_inv = tf.SimilarityTransform(translation=[shift_x, shift_y])

    # This defines the transform you want to perform
    distortion = tf.AffineTransform(matrix=D)

    # skimage transforms can be added like this, does matrix multiplication,
    # hence the need for the brackets. (Note tf.warp takes the inverse)
    transformation = (tf_shift + (distortion + tf_shift_inv)).inverse

    return transformation


def apply_transformation(z, transformation, keep_dtype, order=1, *args, **kwargs):
    """Apply a transformation to a 2-dimensional array.

    Parameters
    ----------
    z : numpy.ndarray
        Array to be transformed
    transformation : numpy.ndarray
        3x3 numpy array specifying the transformation to be applied.
    order : int
        Interpolation order.
    keep_dtype : bool
        If True dtype of returned object is that of z
    *args :
        To be passed to :func:`skimage.transform.warp`
    **kwargs :
        To be passed to :func:`skimage.transform.warp`

    Returns
    -------
    trans : array
        Affine transformed diffraction pattern.

    Notes
    -----
    Generally used in combination with :func:`pyxem.expt_utils.convert_affine_to_transform`

    See Also
    --------
    pyxem.expt_utils.convert_affine_to_transform
    """

    if keep_dtype is False:
        trans = tf.warp(z, transformation, order=order, *args, **kwargs)
    if keep_dtype is True:
        trans = tf.warp(
            z, transformation, order=order, preserve_range=True, *args, **kwargs
        )
        trans = trans.astype(z.dtype)
    return trans


def regional_filter(z, h):
    """Perform a h-dome regional filtering of the an image for background
    subtraction.

    Parameters
    ----------
    h : float
        h-dome cutoff value.

    Returns
    -------
        h-dome subtracted image as numpy.ndarray
    """
    seed = np.copy(z)
    seed = z - h
    mask = z
    dilated = morphology.reconstruction(seed, mask, method="dilation")

    return z - dilated


def circular_mask(shape, radius, center=None):
    """Produces a mask of radius 'r' centered on 'center' of shape 'shape'.

    Parameters
    ----------
    shape : tuple
        The shape of the signal to be masked.
    radius : int
        The radius of the circular mask.
    center : tuple (optional)
        The center of the circular mask. Default: (0, 0)

    Returns
    -------
    mask : numpy.ndarray
        The circular mask.

    """
    l_x, l_y = shape
    x, y = center if center else (l_x / 2, l_y / 2)
    X, Y = np.ogrid[:l_x, :l_y]
    mask = (X - x) ** 2 + (Y - y) ** 2 < radius**2
    return mask


def reference_circle(coords, dimX, dimY, radius):
    """Draw the perimeter of an circle at a given position in the diffraction
    pattern (e.g. to provide a reference for finding the direct beam center).

    Parameters
    ----------
    coords : numpy.ndarray size n,2
        size n,2 array of coordinates to draw the circle.
    dimX : int
        first dimension of the diffraction pattern (size)
    dimY : int
        second dimension of the diffraction pattern (size)
    radius : int
        radius of the circle to be drawn

    Returns
    -------
    img: numpy.ndarray
        Array containing the circle at the position given in the coordinates.
    """
    img = np.zeros((dimX, dimY))

    for n in range(np.size(coords, 0)):
        rr, cc = ellipse_perimeter(coords[n, 0], coords[n, 1], radius, radius)
        img[rr, cc] = 1

    return img


def _find_peak_max(arr, sigma, upsample_factor, kind):
    """Find the index of the pixel corresponding to peak maximum in 1D pattern

    Parameters
    ----------
    sigma : int
        Sigma value for Gaussian blurring kernel for initial beam center estimation.
    upsample_factor : int
        Upsample factor for subpixel maximum finding, i.e. the maximum will
        be found with a precision of 1 / upsample_factor of a pixel.
    kind : str or int, optional
        Specifies the kind of interpolation as a string (‘linear’, ‘nearest’,
        ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’, ‘next’, where
        ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer to a spline
        interpolation of zeroth, first, second or third order; ‘previous’
        and ‘next’ simply return the previous or next value of the point) or as
        an integer specifying the order of the spline interpolator to use.

    Returns
    -------
    center: float
        Pixel position of the maximum
    """
    y1 = ndi.gaussian_filter1d(arr, sigma)
    c1 = np.argmax(y1)  # initial guess for beam center

    m = upsample_factor
    window = 10
    win_len = 2 * window + 1

    try:
        r1 = np.linspace(c1 - window, c1 + window, win_len)
        f = interp1d(r1, y1[c1 - window : c1 + window + 1], kind=kind)
        r2 = np.linspace(
            c1 - window, c1 + window, win_len * m
        )  # extrapolate for subpixel accuracy
        y2 = f(r2)
        c2 = np.argmax(y2) / m  # find beam center with `m` precision
    except ValueError:  # if c1 is too close to the edges, return initial guess
        center = c1
    else:
        center = c2 + c1 - window

    return center


def find_beam_center_interpolate(z, sigma, upsample_factor, kind):
    """Find the center of the primary beam in the image `img` by summing along
    X/Y directions and finding the position along the two directions independently.

    Parameters
    ----------
    sigma : int
        Sigma value for Gaussian blurring kernel for initial beam center estimation.
    upsample_factor : int
        Upsample factor for subpixel beam center finding, i.e. the center will
        be found with a precision of 1 / upsample_factor of a pixel.
    kind : str or int, optional
        Specifies the kind of interpolation as a string (‘linear’, ‘nearest’,
        ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’, ‘next’, where
        ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer to a spline
        interpolation of zeroth, first, second or third order; ‘previous’
        and ‘next’ simply return the previous or next value of the point) or as
        an integer specifying the order of the spline interpolator to use.

    Returns
    -------
    center : numpy.ndarray
        numpy.ndarray, [y, x] containing indices of estimated direct beam positon
    """
    xx = np.sum(z, axis=1)
    yy = np.sum(z, axis=0)

    cx = _find_peak_max(xx, sigma, upsample_factor=upsample_factor, kind=kind)
    cy = _find_peak_max(yy, sigma, upsample_factor=upsample_factor, kind=kind)

    center = np.array([cy, cx])
    return center


def find_beam_center_blur(z, sigma):
    """Estimate direct beam position by blurring the image with a large
    Gaussian kernel and finding the maximum.

    Parameters
    ----------
    sigma : float
        Sigma value for Gaussian blurring kernel.

    Returns
    -------
    center : numpy.ndarray
        numpy.ndarray [x, y] containing indices of estimated direct beam positon.
    """
    if is_cupy_array(z):
        gaus = ndigpu.gaussian_filter
        dispatcher = cp
    else:
        gaus = ndi.gaussian_filter
        dispatcher = np
    blurred = gaus(z, sigma, mode="wrap")
    center = dispatcher.unravel_index(blurred.argmax(), blurred.shape)[::-1]
    return dispatcher.array(center)


def find_center_of_mass(
    signal,
    threshold=None,
    mask=None,
    **kwargs,
):
    if "inplace" in kwargs and kwargs["inplace"]:
        raise ValueError("Inplace is not allowed for center_of_mass")
    else:
        kwargs["inplace"] = False

    det_shape = signal.axes_manager.signal_shape
    if mask is not None:
        x, y, r = mask
        mask = pst._make_circular_mask(x, y, det_shape[0], det_shape[1], r)

    ans = signal.map(
        center_of_mass_from_image,
        threshold=threshold,
        mask=mask,
        **kwargs,
    )
    ans.set_signal_type("beam_shift")
    ans.axes_manager.signal_axes[0].name = "Beam position"
    return ans


def center_of_mass_from_image(z, mask=None, threshold=None):
    """Estimate direct beam position by calculating the center of mass of the
    image.

    Parameters
    ----------
    z : numpy.ndarray
        Two-dimensional data array containing signal.
    mask : numpy.ndarray
        Two-dimensional data array containing mask.
    threshold : float
        Threshold value for center of mass calculation.

    Returns
    -------
    center : numpy.ndarray
        numpy.ndarray [x, y] containing indices of estimated direct beam positon.
    """
    if mask is not None:
        z = z * mask
    if threshold is not None:
        z[z < (np.mean(z) * threshold)] = 0
    center = np.array(ndi.center_of_mass(z))[::-1]
    return center


def find_beam_offset_cross_correlation(z, radius_start, radius_finish, **kwargs):
    """Find the offset of the direct beam from the image center by a cross-correlation algorithm.
    The shift is calculated relative to an circle perimeter. The circle can be
    refined across a range of radii during the centring procedure to improve
    performance in regions where the direct beam size changes,
    e.g. during sample thickness variation.

    Parameters
    ----------
    z: numpy.ndarray
        The two dimensional array/image that is operated on
    radius_start : int
        The lower bound for the radius of the central disc to be used in the
        alignment.
    radius_finish : int
        The upper bounds for the radius of the central disc to be used in the
        alignment.
    **kwargs:
        Any additional keyword arguments defined by :func:`skimage.registration.phase_cross_correlation`

    Returns
    -------
    shift: numpy.ndarray
        numpy.ndarray [y, x] containing offset (from center) of the direct beam positon.
    """
    radiusList = np.arange(radius_start, radius_finish)
    errRecord = np.zeros_like(radiusList, dtype="single")
    origin = np.array(
        [[round(np.size(z, axis=-2) / 2), round(np.size(z, axis=-1) / 2)]]
    )

    for ind in np.arange(0, np.size(radiusList)):
        radius = radiusList[ind]
        ref = reference_circle(origin, np.size(z, axis=-2), np.size(z, axis=-1), radius)
        h0 = np.hanning(np.size(ref, 0))
        h1 = np.hanning(np.size(ref, 1))
        hann2d = np.sqrt(np.outer(h0, h1))
        ref = hann2d * ref
        im = hann2d * z
        shift, error, diffphase = phase_cross_correlation(
            ref, im, upsample_factor=10, **kwargs
        )
        errRecord[ind] = error
        index_min = np.argmin(errRecord)

    ref = reference_circle(
        origin, np.size(z, axis=-2), np.size(z, axis=-1), radiusList[index_min]
    )
    h0 = np.hanning(np.size(ref, 0))
    h1 = np.hanning(np.size(ref, 1))
    hann2d = np.sqrt(np.outer(h0, h1))
    ref = hann2d * ref
    im = hann2d * z
    shift, error, diffphase = phase_cross_correlation(
        ref, im, upsample_factor=100, **kwargs
    )

    shift = shift[::-1]
    return shift - 0.5


def peaks_as_gvectors(z, center, calibration):
    """Converts peaks found as array indices to calibrated units, for use in a
    hyperspy map function.

    Parameters
    ----------
    z : numpy.ndarray
        peak positions as array indices.
    center : numpy.ndarray
        diffraction pattern center in array indices.
    calibration : float
        calibration in reciprocal Angstroms per pixels.

    Returns
    -------
    g : numpy.ndarray
        peak positions in calibrated units.

    """
    g = (z - center) * calibration
    return g


@deprecated(since="0.18.0", removal="1.0.0")
def investigate_dog_background_removal_interactive(
    sample_dp, std_dev_maxs, std_dev_mins
):
    """Utility function to help the parameter selection for the difference of
    gaussians (dog) background subtraction method

    Parameters
    ----------
    sample_dp : ElectronDiffraction2D
        A single diffraction pattern
    std_dev_maxs : iterable
        Linearly spaced maximum standard deviations to be tried, ascending
    std_dev_mins : iterable
        Linearly spaced minimum standard deviations to be tried, ascending

    Returns
    -------
    A hyperspy like navigation (sigma parameters), signal (proccessed patterns)
    plot

    See Also
    --------
    subtract_background_dog : The background subtraction method used.
    numpy.arange : Produces suitable objects for std_dev_maxs

    """
    gauss_processed = np.empty(
        (len(std_dev_maxs), len(std_dev_mins), *sample_dp.axes_manager.signal_shape)
    )

    for i, std_dev_max in enumerate(tqdm(std_dev_maxs, leave=False)):
        for j, std_dev_min in enumerate(std_dev_mins):
            gauss_processed[i, j] = sample_dp.subtract_diffraction_background(
                "difference of gaussians",
                lazy_output=False,
                min_sigma=std_dev_min,
                max_sigma=std_dev_max,
                show_progressbar=False,
            )
    dp_gaussian = pxm.signals.ElectronDiffraction2D(gauss_processed)
    dp_gaussian.metadata.General.title = "Gaussian preprocessed"
    dp_gaussian.axes_manager.navigation_axes[0].name = r"$\sigma_{\mathrm{min}}$"
    dp_gaussian.axes_manager.navigation_axes[1].name = r"$\sigma_{\mathrm{max}}$"
    for axes_number, axes_value_list in [(0, std_dev_mins), (1, std_dev_maxs)]:
        dp_gaussian.axes_manager.navigation_axes[axes_number].offset = axes_value_list[
            0
        ]
        dp_gaussian.axes_manager.navigation_axes[axes_number].scale = (
            axes_value_list[1] - axes_value_list[0]
        )
        dp_gaussian.axes_manager.navigation_axes[axes_number].units = ""

    dp_gaussian.plot(cmap="viridis")
    return None


def find_hot_pixels(z, threshold_multiplier=500, mask=None):
    """Find single pixels which have much larger values compared to neighbors.

    Finds pixels which have a gradient larger than the threshold multiplier.
    These are extremely sharp peaks with values on average much larger than
    the surrounding values.

    Parameters
    ----------
    z : numpy.ndarray
        Frame to operate on
    threshold_multiplier : scaler
        Used to threshold the dif.
    mask : numpy.ndarray, optional
        Array with bool values. The True values will be masked
        (i.e. ignored). Must have the same shape as the two
        last dimensions in dask_array.

    """
    # find the gradient of the image.
    footprint = [[1, 1, 1], [1, 0, 1], [1, 1, 1]]
    median = ndi.median_filter(z, footprint=footprint)
    hot_pixels = (z - median) > threshold_multiplier
    if mask is not None:
        hot_pixels[mask] = False
    return hot_pixels


def remove_bad_pixels(z, bad_pixels):
    """Replace values in bad pixels with mean of neighbors.

    Parameters
    ----------
    z : numpy.ndarray
        A single frame
    bad_pixels : numpy.ndarray
        Must either have the same shape as dask_array,
        or the same shape as the two last dimensions of dask_array.

    Returns
    -------
    data_output : Dask array

    Examples
    --------
    >>> import pyxem.utils.dask_tools as dt
    >>> s = pxm.data.dummy_data.dummy_data.get_dead_pixel_signal(lazy=True)
    >>> dead_pixels = dt._find_dead_pixels(s.data)
    >>> data_output = dt._remove_bad_pixels(s.data, dead_pixels)

    """
    z[bad_pixels] = 0
    bad_pixels_ind = np.transpose(np.array(np.where(bad_pixels)))
    bad_slices = [tuple([slice(i - 1, i + 2) for i in ind]) for ind in bad_pixels_ind]
    values = [np.sum(z[s]) / 8 for s in bad_slices]
    z[bad_pixels] = values
    return z


def normalize_template_match(z, template, subtract_min=True, pad_input=True, **kwargs):
    """Matches a template with an image z. Preformed a normalized cross-correlation
    using the given templates. If subtract_min is True then the minimum value will
    be subtracted from the correlation.

    Parameters
    ----------
    z : numpy.ndarray
        Two-dimensional data array containing signal.
    template : numpy.ndarray
        Two-dimensional data array containing template.
    subtract_min : bool
        If True the minimum value will be subtracted from the correlation.
    pad_input : bool
        If True the input array will be padded. (This should be True otherwise
        the result will be shifted by half the template size)
    **kwargs :
        Keyword arguments to be passed to :func:`skimage.feature.match_template`
    """
    if kwargs.pop("circular_background", False):
        template_match = match_template_dilate(z, template, **kwargs)
    else:
        template_match = match_template(z, template, pad_input=pad_input, **kwargs)
    if subtract_min:
        template_match = template_match - np.min(template_match)
    return template_match


def _supported_float_type(input_dtype, allow_complex=False):
    """Return an appropriate floating-point dtype for a given dtype.

    float32, float64, complex64, complex128 are preserved.
    float16 is promoted to float32.
    complex256 is demoted to complex128.
    Other types are cast to float64.

    Parameters
    ----------
    input_dtype : np.dtype or tuple of np.dtype
        The input dtype. If a tuple of multiple dtypes is provided, each
        dtype is first converted to a supported floating point type and the
        final dtype is then determined by applying `np.result_type` on the
        sequence of supported floating point types.
    allow_complex : bool, optional
        If False, raise a ValueError on complex-valued inputs.

    Returns
    -------
    float_type : dtype
        Floating-point dtype for the image.
    """
    if isinstance(input_dtype, tuple):
        return np.result_type(*(_supported_float_type(d) for d in input_dtype))
    input_dtype = np.dtype(input_dtype)
    if not allow_complex and input_dtype.kind == "c":
        raise ValueError("complex valued input is not supported")
    return new_float_type.get(input_dtype.char, np.float64)


def match_template_dilate(
    image, template, template_dilation=2, mode="constant", constant_values=0
):
    """Matches a template with an image using a window normalized cross-correlation.

    This performs very well for image with different background intensities.  This is a slower version
    of the skimage :func:`skimage.feature.match_template` but performs better for images with circular
    variations in background intensity, specifically accounting for an amorphous halo around the
    diffraction pattern.

    Parameters
    ----------
    image : np.array
        Image to be matched
    template : np.array
        Template to preform the normalized cross-correlation with
    template_dilation : int
        The number of pixels to dilate the template by for the windowed cross-correlation
    mode : str
        Padding mode for the image. Options are 'constant', 'edge', 'wrap', 'reflect'
    constant_values : int
        Value to pad the image with if mode is 'constant'

    Returns
    -------
    response : np.array
        The windowed cross-correlation of the image and template
    """

    image = image.astype(float, copy=False)
    if image.ndim < template.ndim:  # pragma: no cover
        raise ValueError(
            "Dimensionality of template must be less than or "
            "equal to the dimensionality of image."
        )
    if np.any(np.less(image.shape, template.shape)):  # pragma: no cover
        raise ValueError("Image must be larger than template.")

    image_shape = image.shape
    float_dtype = _supported_float_type(image.dtype)

    template = np.pad(template, template_dilation)
    pad_width = tuple((width, width) for width in template.shape)
    if mode == "constant":
        image = np.pad(
            image, pad_width=pad_width, mode=mode, constant_values=constant_values
        )
    else:
        image = np.pad(image, pad_width=pad_width, mode=mode)

    dilated_template = morphology.dilation(
        template, footprint=morphology.disk(template_dilation)
    )
    # Use special case for 2-D images for much better performance in
    # computation of integral images
    image_window_sum = fftconvolve(image, dilated_template[::-1, ::-1], mode="valid")[
        1:-1, 1:-1
    ]
    image_window_sum2 = fftconvolve(
        image**2, dilated_template[::-1, ::-1], mode="valid"
    )[1:-1, 1:-1]

    template_mean = template[
        dilated_template.astype(bool)
    ].mean()  # only consider the pixels in the dilated template
    template_volume = np.sum(dilated_template)
    template_ssd = np.sum((template - template_mean) ** 2)

    xcorr = fftconvolve(image, template[::-1, ::-1], mode="valid")[1:-1, 1:-1]
    numerator = xcorr - image_window_sum * template_mean

    denominator = image_window_sum2
    np.multiply(image_window_sum, image_window_sum, out=image_window_sum)
    np.divide(image_window_sum, template_volume, out=image_window_sum)
    denominator -= image_window_sum
    denominator *= template_ssd
    np.maximum(denominator, 0, out=denominator)  # sqrt of negative number not allowed
    np.sqrt(denominator, out=denominator)

    response = np.zeros_like(xcorr, dtype=float_dtype)

    # avoid zero-division
    mask = denominator > np.finfo(float_dtype).eps

    response[mask] = numerator[mask] / denominator[mask]

    slices = []
    for i in range(template.ndim):
        d0 = (template.shape[i] - 1) // 2
        d1 = d0 + image_shape[i]

        slices.append(slice(d0, d1))

    return response[tuple(slices)]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Classes to help with generating some more complicated analysis/workflow"""

from .calibration_generator import CalibrationGenerator
from .displacement_gradient_tensor_generator import (
    get_DisplacementGradientMap,
    get_single_DisplacementGradientTensor,
)
from .indexation_generator import (
    IndexationGenerator,
    VectorIndexationGenerator,
    TemplateIndexationGenerator,
    ProfileIndexationGenerator,
    AcceleratedIndexationGenerator,
)
from .integration_generator import IntegrationGenerator
from .pdf_generator1d import PDFGenerator1D
from .red_intensity_generator1d import ReducedIntensityGenerator1D
from .subpixelrefinement_generator import SubpixelrefinementGenerator
from .variance_generator import VarianceGenerator
from .virtual_image_generator import VirtualImageGenerator, VirtualDarkFieldGenerator


__all__ = [
    "CalibrationGenerator",
    "get_DisplacementGradientMap",
    "get_single_DisplacementGradientTensor",
    "IndexationGenerator",
    "VectorIndexationGenerator",
    "TemplateIndexationGenerator",
    "ProfileIndexationGenerator",
    "AcceleratedIndexationGenerator",
    "IntegrationGenerator",
    "PDFGenerator1D",
    "ReducedIntensityGenerator1D",
    "SubpixelrefinementGenerator",
    "VarianceGenerator",
    "VirtualImageGenerator",
    "VirtualDarkFieldGenerator",
]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.
from hyperspy.signals import Signal1D
from pyxem.signals import DiffractionVectors
from hyperspy.utils.plot import plot_images


class DiffractionVectors1D(DiffractionVectors, Signal1D):
    """A Collection of Diffraction Vectors with a defined 1D set of vectors.

    For every navigation position there is specifically one vector that represents
    the dataset.

    Examples of DiffractionVectors1D Signals:
        - STEM_DPC
        - STRAIN Maps
        - Diffraction shifts/ Centers

    Attributes
    ----------
    column_scale : np.array()
        The scale for each column in the signal.  For converting the real values
        to pixel values in some image.

    column_offsets : np.array()
        The offsets for each column in the signal.  For converting the real values
        to pixel values in some image.



    """

    _signal_dimension = 1
    _signal_type = "diffraction_vectors"

    def plot(self, tight_layout=True, **kwargs):
        """
        Plot the beam shifts, utilizing HyperSpy's :func:`hyperspy.api.plot.plot_images`
        function. Each Vector is plotted as a separate image with the column names as labels.

        Parameters
        ----------
        tight_layout : bool, optional
            Whether to use tight layout in the plot. The default is True.
        **kwargs : dict
            Keyword arguments to pass to :func:`hyperspy.api.plot.plot_images`.

        """
        if self._lazy:
            raise ValueError(
                "plot is not implemented for lazy signals, " "run compute() first"
            )
        vectors = self.T

        if "suptitle" not in kwargs:
            kwargs["label"] = [c if c is not None else "" for c in self.column_names]
        axes_list = plot_images(
            vectors,
            tight_layout=tight_layout,
            **kwargs,
        )
        return axes_list

    def flatten_diffraction_vectors(self, **kwargs):
        raise NotImplementedError(
            "flatten_diffraction_vectors is not implemented yet for DiffractionVectors1D."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2016-2023 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

exec(open("pyxem/release_info.py").read())  # grab version info

# Projects with optional features for building the documentation and running
# tests. From setuptools:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies
extra_feature_requirements = {
    "doc": [
        "nbsphinx                   >= 0.7",
        "sphinx                     >= 3.0.2",
        "sphinx-copybutton          >= 0.2.5",
        "sphinx-autodoc-typehints   >= 1.10.3",
        "sphinx-gallery             >= 0.6",
        "sphinxcontrib-bibtex       >= 1.0",
        "sphinx_design",
        "sphinx-codeautolink",
        "pydata-sphinx-theme",
        "hyperspy_gui_ipywidgets    >= 2.0",
        "dask-image",
    ],
    "tests": [
        "pytest     >= 5.0",
        "pytest-timeout",
        "pytest-cov >= 2.8.1",
        "pytest-xdist",
        "pytest-rerunfailures",
        "coveralls  >= 1.10",
        "coverage   >= 5.0",
    ],
    "dev": ["black ~= 24.1", "pre-commit >=1.16"],
    "gpu": ["cupy >= 9.0.0"],
    "dask": ["dask-image", "distributed"],
}


setup(
    name=name,
    version=version,
    description="multi-dimensional diffraction microscopy",
    author=author,
    author_email=email,
    license=license,
    url="https://github.com/pyxem/pyxem",
    long_description=open("README.rst").read(),
    keywords=[
        "data analysis",
        "diffraction",
        "microscopy",
        "electron diffraction",
        "electron microscopy",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    packages=find_packages(),
    extras_require=extra_feature_requirements,
    install_requires=[
        "dask",
        "diffsims       >= 0.7.0",  # ensure we use the latest version
        "hyperspy       >= 2.0",
        "h5py",
        "lmfit          >= 0.9.12",
        "matplotlib     >= 3.7.5",
        "numba",
        "numpy",
        "orix           >= 0.12.1",
        "pooch",
        "psutil",
        "scikit-image   >= 0.19.0, !=0.21.0",  # regression in ellipse fitting"
        "scikit-learn   >= 1.0",
        "shapely        > 2.0.0",  # major changes
        "scipy",
        "tqdm",
        "traits",
        "transforms3d",
        "zarr            < 3.0.0",  # for fast saving/loading (>=3.0.0 is not compatible)
    ],
    python_requires=">=3.7",
    package_data={
        "": ["LICENSE", "README.rst"],
        "pyxem": ["*.py", "hyperspy_extension.yaml"],
    },
    entry_points={"hyperspy.extensions": ["pyxem = pyxem"]},
)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from transforms3d.euler import euler2mat

from pyxem.utils.vectors import calculate_norms
from pyxem.utils.vectors import calculate_norms_ragged
from pyxem.utils.vectors import detector_to_fourier
from pyxem.utils.vectors import get_rotation_matrix_between_vectors
from pyxem.utils.vectors import get_angle_cartesian
from pyxem.utils.vectors import get_angle_cartesian_vec
from pyxem.utils.vectors import filter_vectors_near_basis


def test_calculate_norms():
    norms = calculate_norms([[3, 4], [6, 8]])
    assert np.allclose(norms, [5, 10])


def test_calculate_norms_ragged():
    norms = calculate_norms_ragged(np.array([[3], [6, 8]], dtype=object))
    assert np.allclose(norms, [3, 10])


@pytest.mark.parametrize(
    "wavelength, camera_length, detector_coords, k_expected",
    [
        (
            0.025,
            0.2,
            np.array([[0, 0], [0, 1], [1, 1]]),
            np.array(
                [
                    [0, 0, 1 / 0.025 - 40],
                    [0, 1, np.sqrt(1 / (0.025**2) - 1) - 40],
                    [1, 1, np.sqrt(1 / (0.025**2) - 1 - 1) - 40],
                ]
            ),
        )
    ],
)
def test_detector_to_fourier(wavelength, camera_length, detector_coords, k_expected):
    k = detector_to_fourier(detector_coords, wavelength, camera_length)
    np.testing.assert_allclose(k, k_expected)


@pytest.mark.parametrize(
    "from_v1, from_v2, to_v1, to_v2, expected_rotation",
    [
        # v2 from x to y
        (
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0],
            euler2mat(*np.deg2rad([90, 0, 0]), "rzxz"),
        ),
        # Degenerate to-vectors gives half-way rotation (about y-axis)
        (
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [2.0, 0.0, 0.0],
            euler2mat(*np.deg2rad([90, 45, -90]), "rzxz"),
        ),
        # Edges to body diagonals
        (
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.5, -0.5, 1 / np.sqrt(2)],
            [1 / np.sqrt(2), 1 / np.sqrt(2), 0],
            euler2mat(*np.deg2rad([45, 45, 0]), "rzxz"),
        ),
    ],
)
def test_get_rotation_matrix_between_vectors(
    from_v1, from_v2, to_v1, to_v2, expected_rotation
):
    rotation_matrix = get_rotation_matrix_between_vectors(
        np.array(from_v1), np.array(from_v2), np.array([to_v1]), np.array([to_v2])
    )
    np.testing.assert_allclose(
        rotation_matrix, np.array([expected_rotation]), atol=1e-15
    )


@pytest.mark.parametrize(
    "vec_a, vec_b, expected_angle",
    [([0, 0, 1], [0, 1, 0], np.deg2rad(90)), ([0, 0, 0], [0, 0, 1], 0)],
)
def test_get_angle_cartesian(vec_a, vec_b, expected_angle):
    angle = get_angle_cartesian(vec_a, vec_b)
    np.testing.assert_allclose(angle, expected_angle)


@pytest.mark.parametrize(
    "a, b, expected_angles",
    [
        (
            np.array([[0, 0, 1], [0, 0, 0]]),
            np.array([[0, 1, 0], [0, 0, 1]]),
            [np.deg2rad(90), 0],
        )
    ],
)
def test_get_angle_cartesian_vec(a, b, expected_angles):
    angles = get_angle_cartesian_vec(a, b)
    np.testing.assert_allclose(angles, expected_angles)


# @pytest.mark.xfail(raises=ValueError)
def test_get_angle_cartesian_vec_input_validation():
    # Note - uses regex via re.search()
    with pytest.raises(
        ValueError,
        match=r"shape of a .* and b .* must be the same",
    ):
        get_angle_cartesian_vec(np.empty((2, 3)), np.empty((5, 3)))


def test_filter_near_basis():
    rng = np.random.default_rng(10)
    basis_vectors = rng.integers(0, 100, (10, 2))
    shifts = rng.random((10, 2))
    dist = np.linalg.norm(shifts, axis=1)
    shifted = basis_vectors + shifts
    filt = filter_vectors_near_basis(shifted, basis_vectors, distance=0.4)
    is_nan = np.isnan(filt).sum(axis=1) > 0
    np.testing.assert_array_equal(is_nan, dist > 0.4)


def test_basis_filter_no_vectors():
    basis_vectors = np.random.randint(0, 100, (10, 2))
    vectors = np.empty((0, 2))
    filt = filter_vectors_near_basis(vectors, basis_vectors, distance=0.4)
    is_nan = np.isnan(filt).sum(axis=1) > 0
    np.testing.assert_array_equal(is_nan, True)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Additional utility functions for processing signals.

.. currentmodule:: pyxem.utils

.. rubric:: Modules

.. autosummary::
    :toctree: ../generated/
    :template: custom-module-template.rst

    plotting
    ransac_ellipse_tools
    vectors
    diffraction
    calibration

"""


from pyxem.utils.ransac_ellipse_tools import determine_ellipse
from pyxem.utils.calibration import find_diffraction_calibration
from pyxem.utils.plotting import plot_template_over_pattern

from pyxem.utils import vectors
from pyxem.utils import plotting
from pyxem.utils import ransac_ellipse_tools
from pyxem.utils import calibration
from pyxem.utils import diffraction

__all__ = [
    "find_diffraction_calibration",
    "plot_template_over_pattern",
    "determine_ellipse",
    "ransac_ellipse_tools",
    "vectors",
    "calibration",
    "plotting",
    "diffraction",
]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Generating DisplacementGradientMaps from diffraction vectors."""

import numpy as np
from pyxem.signals.tensor_field import DisplacementGradientMap


def get_DisplacementGradientMap(
    strained_vectors,
    unstrained_vectors,
    weights=None,
    return_residuals=False,
    columns=[0, 1],
    **kwargs
):
    r"""Calculates the displacement gradient tensor at each navigation position in a map.

    Compares vectors to determine the 2 x 2 matrix,
    :math:`\\mathbf(L)`, that maps unstrained vectors, Vu, to strained vectors,
    Vs, using the np.lingalg.inv() function to find L that satisfies
    :math:`Vs = \\mathbf(L) Vu`.

    The transformation is returned as a 3 x 3 displacement gradient tensor.

    Parameters
    ----------
    strained_vectors : hyperspy.Signal2D
        Signal2D with a 2 x n array at each navigation position containing the
        Cartesian components of two strained basis vectors, V and U, defined as
        row vectors.
    unstrained_vectors : numpy.array
        A 2 x n array containing the Cartesian components of two unstrained
        basis vectors, V and U, defined as row vectors.
    weights : list
        of weights to be passed to the least squares optimiser, not used for n=2
    return_residuals: Bool
        If the residuals for the least squares optimiser should be returned.
    kwargs: dict
        Any additional keyword arguments passed to the `hyperspy.signals.BaseSignal.map`
        function.

    Returns
    -------
    D : DisplacementGradientMap
        The 3 x 3 displacement gradient tensor (measured in reciprocal space) at
        every navigation position.

    Notes
    -----
    n=2 now behaves the same as the n>2 case; see Release Notes for 0.10.0 for details.

    See Also
    --------
    get_single_DisplacementGradientTensor()

    """

    # Calculate displacement gradient tensor across map.
    D = strained_vectors.map(
        get_single_DisplacementGradientTensor,
        Vu=unstrained_vectors,
        weights=weights,
        inplace=False,
        output_signal_size=(3, 3),
        ragged=False,
        output_dtype=np.float64,
        columns=columns,
        **kwargs
    )
    D.set_signal_type("tensor_field")
    if return_residuals:
        R = strained_vectors.map(
            get_single_DisplacementGradientTensor,
            Vu=unstrained_vectors,
            weights=weights,
            inplace=False,
            output_dtype=np.float64,
            return_residuals=True,
            columns=columns,
            **kwargs
        )
        return D, R
    else:
        return D


def get_single_DisplacementGradientTensor(
    Vs, Vu=None, weights=None, return_residuals=False, columns=[0, 1]
):
    r"""Calculates the displacement gradient tensor from a pairs of vectors.

    Determines the 2 x 2 matrix, :math:`\\mathbf(L)`, that maps unstrained
    vectors, Vu, onto strained vectors, Vs

    The transformation is returned as a 3 x 3 displacement gradient tensor.

    Parameters
    ----------
    Vs : numpy.array
        A 2 x n array containing the Cartesian components of two strained basis
        vectors, V and U, defined as row vectors.
    Vu : numpy.array
        A 2 x n array containing the Cartesian components of two unstrained
        basis vectors, V and U, defined as row vectors.
    weights : list
        of weights to be passed to the least squares optimiser
    return_residuals: Bool
        If the residuals for the least squares optimiser should be returned.
    Returns
    -------
    D : numpy.array
        A 3 x 3 displacement gradient tensor (measured in reciprocal space).
    residuals : numpy.array
        The residuals for the least squares fitting.

    Notes
    -----
    n=2 now behaves the same as the n>2 case; see Release Notes for 0.10.0 for details.

    See Also
    --------
    get_DisplacementGradientMap()

    """
    Vu = Vu[:, columns]
    Vs = Vs[:, columns]
    is_row_nan = np.logical_not(np.any(np.isnan(Vs), axis=1))
    Vs = Vs[is_row_nan]
    Vu = Vu[is_row_nan]

    if Vu is not None:
        if Vu.dtype == object:
            Vu = Vu[()]

    if weights is not None:
        # see https://stackoverflow.com/questions/27128688
        weights = np.asarray(weights)
        # Need vectors normalized to the unstrained region otherwise the weighting breaks down
        Vs = (np.divide(Vs.T, np.linalg.norm(Vu, axis=1)) * np.sqrt(weights)).T
        Vu = (np.divide(Vu.T, np.linalg.norm(Vu, axis=1)) * np.sqrt(weights)).T
    else:
        Vs, Vu = Vs, Vu

    L, residuals, rank, s = np.linalg.lstsq(Vu, Vs, rcond=-1)
    # only need the return array, see np,linalg.lstsq doc
    # Put caculated matrix values into 3 x 3 matrix to be returned.
    D = np.eye(3)
    D[0:2, 0:2] = L

    if return_residuals:
        return residuals
    else:
        return D



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from hyperspy.signals import Signal2D

from pyxem.signals import (
    ElectronDiffraction1D,
    ElectronDiffraction2D,
    PolarDiffraction2D,
    LazyElectronDiffraction2D,
)


def test_init():
    z = np.zeros((2, 2, 2, 2))
    _ = ElectronDiffraction2D(
        z, metadata={"Acquisition_instrument": {"SEM": "Expensive-SEM"}}
    )


class TestSimpleMaps:
    # Confirms that maps run without error.

    def test_center_direct_beam_cross_correlate(self, diffraction_pattern):
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)
        diffraction_pattern.center_direct_beam(
            method="cross_correlate", radius_start=1, radius_finish=3
        )
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)

    @pytest.mark.skip(reason="Broken downstream in scikit image")
    def test_center_direct_beam_xc_return_shifts(self, diffraction_pattern):
        shifts = diffraction_pattern.center_direct_beam(
            method="cross_correlate",
            radius_start=1,
            radius_finish=3,
            return_shifts=True,
        )
        ans = np.array([[[0.45, 0.45], [-0.57, -0.57]], [[0.45, 0.45], [-0.52, -0.52]]])
        np.testing.assert_almost_equal(shifts, ans)

    def test_center_direct_beam_blur_return_shifts(self, diffraction_pattern):
        shifts = diffraction_pattern.center_direct_beam(
            method="blur", sigma=5, half_square_width=3, return_shifts=True
        )
        ans = np.array([[[1.0, 1.0], [0.0, 0.0]], [[1.0, 1.0], [0.0, 0.0]]])
        np.testing.assert_almost_equal(shifts, ans)

    def test_center_direct_beam_in_small_region(self, diffraction_pattern):
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)
        diffraction_pattern.center_direct_beam(
            method="blur", sigma=int(5), half_square_width=3
        )
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)

    def test_apply_affine_transformation(self, diffraction_pattern):
        diffraction_pattern.apply_affine_transformation(
            D=np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        )
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)

    def test_apply_affine_transformation_upsample(self, diffraction_pattern):
        old_data = np.copy(diffraction_pattern.data)
        diffraction_pattern.apply_affine_transformation(
            D=np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
            output_shape=(8, 8),
        )
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)
        assert diffraction_pattern.axes_manager.signal_shape == (8, 8)
        np.testing.assert_almost_equal(old_data, diffraction_pattern.data)

    def test_apply_affine_transforms_paths(self, diffraction_pattern):
        D = np.array([[1.0, 0.9, 0.0], [1.1, 1.0, 0.0], [0.0, 0.0, 1.0]])
        s = Signal2D(np.asarray([[D, D], [D, D]]))
        static = diffraction_pattern.apply_affine_transformation(D, inplace=False)
        dynamic = diffraction_pattern.apply_affine_transformation(s, inplace=False)
        assert np.allclose(static.data, dynamic.data, atol=1e-3)

    def test_apply_affine_transformation_with_casting(self, diffraction_pattern):
        diffraction_pattern.change_dtype("uint8")

        with pytest.warns(
            UserWarning,
            match="Bi-quadratic interpolation behavior has changed due to a bug in the implementation of scikit-image",
        ):
            transformed_dp = ElectronDiffraction2D(
                diffraction_pattern
            ).apply_affine_transformation(
                D=np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.2]]),
                order=2,
                keep_dtype=True,
                inplace=False,
            )

        assert transformed_dp.data.dtype == "uint8"


class TestSimpleHyperspy:
    # Tests functions that assign to hyperspy metadata

    def test_set_experimental_parameters(self, diffraction_pattern):
        diffraction_pattern.set_experimental_parameters(
            beam_energy=3.0,
            camera_length=3.0,
            scan_rotation=1.0,
            convergence_angle=1.0,
            rocking_angle=1.0,
            rocking_frequency=1.0,
            exposure_time=1.0,
        )
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)

    def test_set_scan_calibration(self, diffraction_pattern):
        diffraction_pattern.set_scan_calibration(19)
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)

    @pytest.mark.parametrize(
        "calibration, center",
        [
            (
                1,
                (4, 4),
            ),
            (0.017, (3, 3)),
            (
                0.5,
                None,
            ),
        ],
    )
    def test_set_diffraction_calibration(
        self, diffraction_pattern, calibration, center
    ):
        calibrated_center = (
            calibration * np.array(center) if center is not None else center
        )
        diffraction_pattern.set_diffraction_calibration(
            calibration, center=calibrated_center
        )
        dx, dy = diffraction_pattern.axes_manager.signal_axes
        assert dx.scale == calibration and dy.scale == calibration
        if center is not None:
            assert np.all(
                diffraction_pattern.isig[0.0, 0.0].data
                == diffraction_pattern.isig[center[0], center[1]].data
            )


class TestGainNormalisation:
    @pytest.mark.parametrize(
        "dark_reference, bright_reference",
        [
            (-1, 1),
            (0, 1),
            (0, 256),
        ],
    )
    def test_apply_gain_normalisation(
        self, diffraction_pattern, dark_reference, bright_reference
    ):
        dpr = diffraction_pattern.apply_gain_normalisation(
            dark_reference=dark_reference,
            bright_reference=bright_reference,
            inplace=False,
        )
        assert dpr.max() == bright_reference
        assert dpr.min() == dark_reference


class TestDirectBeamMethods:
    @pytest.mark.parametrize(
        "mask_expected",
        [
            (
                np.array(
                    [
                        [False, False, False, False, False, False, False, False],
                        [False, False, False, False, False, False, False, False],
                        [False, False, False, True, True, False, False, False],
                        [False, False, True, True, True, True, False, False],
                        [False, False, True, True, True, True, False, False],
                        [False, False, False, True, True, False, False, False],
                        [False, False, False, False, False, False, False, False],
                        [False, False, False, False, False, False, False, False],
                    ]
                ),
            ),
        ],
    )
    def test_get_direct_beam_mask(self, diffraction_pattern, mask_expected):
        mask_calculated = diffraction_pattern.get_direct_beam_mask(2)
        assert isinstance(mask_calculated, Signal2D)
        assert np.equal(mask_calculated, mask_expected)


class TestComputeAndAsLazyElectron2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyElectronDiffraction2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert isinstance(s, ElectronDiffraction2D)
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyElectronDiffraction2D(dask_array)
        s.compute()
        assert s.__class__ == ElectronDiffraction2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = ElectronDiffraction2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyElectronDiffraction2D)
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = ElectronDiffraction2D(data)
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyElectronDiffraction2D)
        assert data.shape == s_lazy.data.shape


class TestDecomposition:
    def test_decomposition_class_assignment(self, diffraction_pattern):
        diffraction_pattern.decomposition()
        assert isinstance(diffraction_pattern, ElectronDiffraction2D)


class TestIntegration:
    @pytest.fixture
    def ones(self):
        ones_diff = ElectronDiffraction2D(data=np.ones(shape=(5, 5)))
        ones_diff.axes_manager.signal_axes[0].name = "kx"
        ones_diff.axes_manager.signal_axes[1].name = "ky"
        ones_diff.unit = "2th_rad"
        return ones_diff

    @pytest.mark.parametrize("energy", [None, 200])
    def test_1d_azimuthal_integration(self, ones, energy):
        ones.beam_energy = energy
        integration = ones.get_azimuthal_integral1d(npt=10)
        assert isinstance(integration, ElectronDiffraction1D)

    @pytest.mark.parametrize("energy", [None, 200])
    def test_2d_azimuthal_integration(self, ones, energy):
        ones.beam_energy = energy
        integration = ones.get_azimuthal_integral2d(npt=10)
        assert isinstance(integration, PolarDiffraction2D)

    def test_set_scan_calibration(self):
        ones = ElectronDiffraction2D(data=np.ones((3, 3, 3, 3)))
        ones.scan_calibration = 0.9
        assert ones.axes_manager.navigation_axes[0].scale == 0.9
        assert ones.axes_manager.navigation_axes[1].scale == 0.9
        assert ones.scan_calibration == 0.9

    def test_set_diffraction_calibration(self):
        ones = ElectronDiffraction2D(data=np.ones((3, 3, 3, 3)))
        ones.diffraction_calibration = 0.9
        assert ones.axes_manager.signal_axes[0].scale == 0.9
        assert ones.axes_manager.signal_axes[1].scale == 0.9
        assert ones.diffraction_calibration == 0.9

    def test_set_camera_length(self, ones):
        assert ones.camera_length is None
        ones.camera_length = 1.5
        assert ones.metadata.Acquisition_instrument.TEM["camera_length"] == 1.5
        assert ones.camera_length == 1.5

    def test_set_beam_energy(self, ones):
        assert ones.beam_energy is None
        ones.beam_energy = 1.5
        assert ones.metadata.Acquisition_instrument.TEM["beam_energy"] == 1.5
        assert ones.beam_energy == 1.5



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.optimize import curve_fit

from hyperspy._signals.lazy import LazySignal
import hyperspy.api as hs
from diffsims.utils.sim_utils import get_electron_wavelength
from diffsims.simulations.simulation1d import Simulation1D

from pyxem.signals.diffraction1d import Diffraction1D


class ElectronDiffraction1D(Diffraction1D):
    """Signal class for Electron Diffraction radial profiles.

    Parameters
    ----------
    *args
        See :class:`hyperspy.api.signals.Signal1D`.
    **kwargs
        See :class:`hyperspy.api.signals.Signal1D`
    """

    _signal_type = "electron_diffraction"

    from scipy.optimize import curve_fit

    def model_simulation1d(
        self,
        simulation,
        power_law_background=True,
        starting_scale=None,
        auto_limit_peaks=True,
        center_lim=0.05,
        fit=True,
    ):
        """Creates a model for fitting diffraction peaks for a ring pattern using a series of Gaussians.

        This is primarily useful for calibration but could be potentially used for other things like
        phase segmentation from 1D diffraction patterns.  The lack of a parallel option for model fitting
        in hyperspy limits that usefulness for 4D STEM.  If that is something you are interested in doing,
         raise an issue for more discussion.

        Parameters
        ----------
        simulation : diffsims.simulations.Simulation1D
            The simulation of a 1D "powder" diffraction pattern.
        power_law_background : bool
            Whether to include a power law background in the model.
        starting_scale : float
            The starting scale for the model.  If None, the maximum peak intensity in the simulation is used.
            Note that removing the zero peak is generally a good idea to help with fitting.
        auto_limit_peaks : bool
            Whether to automatically guess some reasonable parameters for the Gaussian peaks and bounds.
        fit : bool
            Whether to fit the model.  If False, the model is returned unfitted.

        Returns
        -------
        model : hyperspy.models.Model1D
            The model for fitting the diffraction peaks to a 1D diffraction pattern.

        Notes
        -----
        This functionality is still under development and may change slightly in the future.

        """
        if starting_scale == None:
            max_peak_inten_sim = simulation.reciprocal_spacing[
                np.argmax(simulation.intensities)
            ]
            max_peak_data = self.axes_manager.signal_axes[0].axis[np.argmax(self.data)]
            starting_scale = max_peak_inten_sim / max_peak_data
        model = self.create_model()

        if power_law_background:
            model.append(hs.model.components1D.PowerLaw())
        for r in simulation.reciprocal_spacing:
            g = hs.model.components1D.Gaussian(centre=r / starting_scale)
            if (
                auto_limit_peaks
            ):  # guessing some reasonable parameters for "typical" diffraction rings
                g.centre.bmin = r / starting_scale - center_lim / starting_scale
                g.centre.bmax = r / starting_scale + center_lim / starting_scale
                g.sigma.bmax = 0.03 / starting_scale
                g.sigma.bmin = 0.01 / starting_scale
            model.append(g)
        if fit:
            model.fit(bounded=True)
        return model

    def model2theta_scale(self, simulation, model, beam_energy):
        """Find the theta calibration scale for a 1d diffraction pattern based
        on a :class:`diffsims.simulations.Simulation1D`, a :class:`hyperspy.models.Model1D`
        and the beam energy.

        Parameters
        ----------
        simulation : diffsims.simulations.Simulation1D
            The simulation of a 1D "powder" diffraction pattern.
        model : hyperspy.models.Model1D
            The model for fitting the diffraction peaks to a 1D diffraction pattern.
        beam_energy : float
            The beam energy in keV.

        Notes
        -----
        This functionality is still under development and may change slightly in the future.
        """

        def f(x, m):
            return x * m

        centers = np.array(
            [
                c.centre.value
                for c in model
                if isinstance(c, hs.model.components1D.Gaussian)
            ]
        )
        wavelength = get_electron_wavelength(beam_energy)
        angles = np.arctan2(np.sort(simulation.reciprocal_spacing), 1 / wavelength)
        m, pcov = curve_fit(f, np.sort(centers), angles)
        scale = m[0]
        return scale

    def model2camera_length(self, simulation, model, beam_energy, physical_pixel_size):
        """Calculate the camera length from a model and simulation.

        This is useful for calibrating the camera length very exactly.

        Parameters
        ----------
        simulation : diffsims.simulations.Simulation1D
            The simulation of a 1D "powder" diffraction pattern.
        model : hyperspy.models.Model1D
            The model for fitting the diffraction peaks to a 1D diffraction pattern.
        beam_energy : float
            The beam energy in keV.
        physical_pixel_size : float
            The physical pixel size in meters. Note the Merlin has a pixel size of 55 um
            or 5.5E-5 m and the Celeritas/ Celeritas XS has a 15 um pixel size or 1.5E-5 m.

        Returns
        -------
        camera_length : float
            The camera length in meters for the experiment.  Note that this is often
            slightly different from the micrscope setting, however this calibrated
            value is more accurate than the calculated value from the micrscope.

        Notes
        -----
        This functionality is still under development and may change slightly in the future.
        """
        theta_scale = self.model2theta_scale(simulation, model, beam_energy)
        camera_length = physical_pixel_size / np.tan(theta_scale)
        return camera_length

    def set_experimental_parameters(
        self,
        accelerating_voltage=None,
        camera_length=None,
        scan_rotation=None,
        convergence_angle=None,
        rocking_angle=None,
        rocking_frequency=None,
        exposure_time=None,
    ):
        """Set experimental parameters in metadata.

        Parameters
        ----------
        accelerating_voltage : float
            Accelerating voltage in kV
        camera_length: float
            Camera length in cm
        scan_rotation : float
            Scan rotation in degrees
        convergence_angle : float
            Convergence angle in mrad
        rocking_angle : float
            Beam rocking angle in mrad
        rocking_frequency : float
            Beam rocking frequency in Hz
        exposure_time : float
            Exposure time in ms.
        """
        md = self.metadata

        if accelerating_voltage is not None:
            md.set_item(
                "Acquisition_instrument.TEM.accelerating_voltage", accelerating_voltage
            )
        if camera_length is not None:
            md.set_item(
                "Acquisition_instrument.TEM.Detector.Diffraction.camera_length",
                camera_length,
            )
        if scan_rotation is not None:
            md.set_item("Acquisition_instrument.TEM.scan_rotation", scan_rotation)
        if convergence_angle is not None:
            md.set_item(
                "Acquisition_instrument.TEM.convergence_angle", convergence_angle
            )
        if rocking_angle is not None:
            md.set_item("Acquisition_instrument.TEM.rocking_angle", rocking_angle)
        if rocking_frequency is not None:
            md.set_item(
                "Acquisition_instrument.TEM.rocking_frequency", rocking_frequency
            )
        if exposure_time is not None:
            md.set_item(
                "Acquisition_instrument.TEM.Detector.Diffraction.exposure_time",
                exposure_time,
            )

    def set_diffraction_calibration(self, calibration):
        """Set diffraction profile channel size in reciprocal Angstroms.

        Parameters
        ----------
        calibration : float
            Diffraction profile calibration in reciprocal Angstroms per pixel.
        """
        dx = self.axes_manager.signal_axes[0]

        dx.name = "k"
        dx.scale = calibration
        dx.units = "$A^{-1}$"

    def set_scan_calibration(self, calibration):
        """Set scan pixel size in nanometres.

        Parameters
        ----------
        calibration: float
            Scan calibration in nanometres per pixel.
        """
        x = self.axes_manager.navigation_axes[0]
        y = self.axes_manager.navigation_axes[1]

        x.name = "x"
        x.scale = calibration
        x.units = "nm"

        y.name = "y"
        y.scale = calibration
        y.units = "nm"


class LazyElectronDiffraction1D(LazySignal, ElectronDiffraction1D):
    pass



from pyxem.data.dummy_data import CrystalSTEMSimulation
from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D
from pyxem.data.simulated_fe import fe_fcc_phase


def simulated_overlap(
    real_space_pixels=128,
    num_crystals=50,
    noise_level=0.1,
    num_electrons=10,
    radius=5,
    recip_space_pixels=64,
    k_range=1,
):
    """
    Create a simulated diffraction pattern with overlapping nanocrystals.

    Returns
    -------
    diffraction_pattern : Signal2D
        A simulated diffraction pattern with overlapping nanocrystals.
    """
    fcc_fe = fe_fcc_phase()
    sim = CrystalSTEMSimulation(
        fcc_fe,
        real_space_pixels=real_space_pixels,
        num_crystals=num_crystals,
        recip_space_pixels=recip_space_pixels,
        k_range=k_range,
    )
    arr = sim.make_4d_stem(
        num_electrons=num_electrons, noise_level=noise_level, radius=radius
    )
    signal = ElectronDiffraction2D(arr)
    signal.axes_manager.signal_axes[1].scale = k_range / (recip_space_pixels / 2)
    signal.axes_manager.signal_axes[0].scale = k_range / (recip_space_pixels / 2)
    signal.calibration.center = None
    return signal



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import hyperspy.api as hs

from pyxem.generators import VirtualImageGenerator, VirtualDarkFieldGenerator
from pyxem.signals import (
    Diffraction2D,
    DiffractionVectors2D,
    ElectronDiffraction2D,
    DiffractionVectors,
    VirtualDarkFieldImage,
)


@pytest.fixture(params=[np.array([[1, 1], [2, 2]])])
def diffraction_vectors(request):
    dvec = DiffractionVectors2D(request.param)
    return dvec


@pytest.fixture
def vdf_generator(diffraction_pattern, diffraction_vectors):
    diffraction_pattern.data = np.where(
        diffraction_pattern.data == 0, 0.01, diffraction_pattern.data
    )  # avoid divide by zeroes
    return VirtualDarkFieldGenerator(diffraction_pattern, diffraction_vectors)


class TestVirtualDarkFieldGenerator:
    def setup_method(self):
        object1 = np.zeros((5, 5), dtype=bool)
        object1[2:4, 2:4] = True
        self.object1 = object1

        object2 = np.zeros((5, 5), dtype=bool)
        object2[1:3, 3:5] = True
        self.object2 = object2

        data = np.zeros((5, 5, 10, 10))

        data[object1, 6, 7] = 7

        data[object1, 2, 3] = 8

        data[object2, 1, 4] = 9

        data[object2, 4, 7] = 10

        self.vectors = DiffractionVectors2D([[6, 7], [2, 3], [1, 4], [4, 7]])

        self.data = Diffraction2D(data)

        self.vdf = VirtualDarkFieldGenerator(self.data, self.vectors)

    def test_setup(self):
        assert isinstance(self.vdf, VirtualDarkFieldGenerator)
        assert isinstance(self.data, Diffraction2D)
        assert isinstance(self.vectors, DiffractionVectors2D)

    @pytest.mark.parametrize("radius, normalize", [(1.0, False), (1.0, True)])
    def test_get_virtual_dark_field_images2(self, vdf_generator, radius, normalize):
        vdfs = self.vdf.get_virtual_dark_field_images(radius, normalize)
        assert isinstance(vdfs, VirtualDarkFieldImage)
        for i, value in enumerate([7, 8, 9, 10]):
            if normalize:
                assert np.sum(vdfs.data[i]) == 4
            else:
                assert np.sum(vdfs.data[i]) == value * np.sum(self.object1)

    def test_vdf_generator_init_with_vectors(self, diffraction_pattern):
        dvm = DiffractionVectors(
            np.array(
                [
                    [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2], [1, 2]])],
                    [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2]])],
                ],
                dtype=object,
            )
        )
        assert isinstance(dvm, DiffractionVectors)
        vdfgen = VirtualDarkFieldGenerator(diffraction_pattern, dvm)

        assert isinstance(vdfgen.signal, ElectronDiffraction2D)
        assert isinstance(vdfgen.vectors, DiffractionVectors2D)

    @pytest.mark.parametrize("radius, normalize", [(4.0, False), (4.0, True)])
    def test_get_virtual_dark_field_images(self, vdf_generator, radius, normalize):
        vdfs = vdf_generator.get_virtual_dark_field_images(radius, normalize)
        assert isinstance(vdfs, VirtualDarkFieldImage)


class TestVirtualImageGenerator:
    def setup_method(self, method):
        # Navigation dimension of the diffraction patterns
        diffraction_pattern = Diffraction2D(np.arange(2000).reshape(4, 5, 10, 10))
        virtual_image_generator = VirtualImageGenerator(diffraction_pattern)
        self.virtual_image_generator = virtual_image_generator
        self.diffraction_pattern = diffraction_pattern

    @pytest.mark.parametrize(
        "k_min, k_max, k_steps, normalize", [(0.0, 4.0, 2, False), (0.0, 4.0, 2, True)]
    )
    def test_get_concentric_virtual_images(self, k_min, k_max, k_steps, normalize):
        vdfs = self.virtual_image_generator.get_concentric_virtual_images(
            k_min, k_max, k_steps, normalize
        )
        assert isinstance(vdfs, VirtualDarkFieldImage)

    def test_calibration_vdf_images(self):
        diffraction_pattern = self.diffraction_pattern
        virtual_image_generator = self.virtual_image_generator
        nav_axis = diffraction_pattern.axes_manager.navigation_axes[0]
        nav_axis.scale = 0.2
        nav_axis.offset = 10
        nav_axis.units = "nm"
        nav_axis.name = "position"
        for sig_axis in diffraction_pattern.axes_manager.signal_axes:
            sig_axis.scale = 0.2
            sig_axis.offset = -1.0
            sig_axis.units = "rad"
            sig_axis.name = "Scattering Angle"
        k_min, k_max, k_steps = 0.1, 0.6, 2
        vi = virtual_image_generator.get_concentric_virtual_images(
            k_min, k_max, k_steps
        )

        assert vi.data.shape == (2, 4, 5)

        vi_nav_axis = vi.axes_manager[0]
        assert vi_nav_axis.name == "Annular bins"
        assert vi_nav_axis.units == sig_axis.units
        assert vi_nav_axis.scale == (k_max - k_min) / k_steps
        assert vi_nav_axis.offset == k_min
        assert vi_nav_axis.size == k_steps

        vi_sig_axis = vi.axes_manager[1]
        for attr in ["scale", "offset", "units", "name"]:
            assert getattr(vi_sig_axis, attr) == getattr(nav_axis, attr)


def test_vdf_generator_from_map(diffraction_pattern):
    dvm = DiffractionVectors(
        np.array(
            [
                [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2], [1, 2]])],
                [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2]])],
            ],
            dtype=object,
        )
    )
    vdfgen = VirtualImageGenerator(diffraction_pattern, dvm)
    assert isinstance(vdfgen, VirtualImageGenerator)


def test_vi_generator_set_ROI_mesh(diffraction_pattern):
    vi_generator = VirtualImageGenerator(diffraction_pattern)
    diffraction_pattern.plot()
    vi_generator.set_ROI_mesh(1.0, 1.2, 1.0)
    assert len(vi_generator.roi_list) == 3
    assert isinstance(vi_generator.roi_list[0], hs.roi.CircleROI)


@pytest.mark.parametrize("nav_shape", [[2], [2, 4]])
def test_vi_generator_get_virtual_images_from_mesh(nav_shape):
    n = np.prod(nav_shape) * 32**2
    s = Diffraction2D(np.arange(n).reshape((*nav_shape, 32, 32)))
    vi_generator = VirtualImageGenerator(s)

    with pytest.raises(ValueError):
        vi_generator.get_virtual_images_from_mesh()

    for axis in s.axes_manager.signal_axes:
        axis.scale = 0.1
        axis.offset = -1.6
        axis.units = "1/nm"

    vi_generator.set_ROI_mesh(0.5, 0.6, 1.4)
    vi = vi_generator.get_virtual_images_from_mesh()
    vi_nav_axis = vi.axes_manager.navigation_axes[0]
    assert vi_nav_axis.size == 21
    assert vi_nav_axis.name == "ROI index"
    assert vi_nav_axis.scale == 1.0
    assert len(vi_generator.roi_list) == 21

    vi_generator.set_ROI_mesh(0.5, 0.6, 1.0)
    assert len(vi_generator.roi_list) == 11

    vi = vi_generator.get_virtual_images_from_mesh(normalize=True)


def test_vi_generator_get_virtual_images_from_mesh_nav_dim3():
    s = Diffraction2D(np.arange(2 * 4 * 10 * 32**2).reshape((2, 4, 10, 32, 32)))
    for axis in s.axes_manager.signal_axes:
        axis.scale = 0.1
        axis.offset = -1.6
        axis.units = "1/nm"
    vi_generator = VirtualImageGenerator(s)

    vi_generator.set_ROI_mesh(0.5, 0.6, 0.6)
    vi = vi_generator.get_virtual_images_from_mesh()

    assert vi.axes_manager.navigation_shape == (2, 5)

    vi = vi_generator.get_virtual_images_from_mesh(out_signal_axes=[1, 2])

    assert vi.axes_manager.navigation_shape == (10, 5)



import pyxem as pxm
from skimage.draw import disk
import numpy as np
from scipy.ndimage import gaussian_filter


def simulated_stripes(beam_shifts=None):
    """
    Create a simulated diffraction pattern with stripes for showing magnetic DPC.

    Parameters
    ----------
    shifts : array_like, optional
        The shifts to apply to the direct beam.  The default is None which
        corresponds to a shift of [[3, 3], [-3,3]] for the alternating stripes.
    Returns
    -------
    diffraction_pattern : Signal2D
        A simulated diffraction pattern with the direct beam shifted by .
    """
    if beam_shifts is None:
        beam_shifts = np.array([[3, 3], [-3, 3]])
    data = np.zeros((60, 60, 64, 64))
    rr, cc = disk(center=(32, 32), radius=16)
    data[:, :, rr, cc] = 1
    s = pxm.signals.ElectronDiffraction2D(data)
    shifts = np.zeros((60, 60, 2))
    shifts[20:40, 10:20] = beam_shifts[0]
    shifts[20:40, 20:30] = beam_shifts[1]
    shifts[20:40, 30:40] = beam_shifts[0]
    shifts[20:40, 40:50] = beam_shifts[1]
    bs = pxm.signals.BeamShift(shifts)
    s.center_direct_beam(shifts=bs)  # Shift the direct beam
    s.axes_manager[2].scale = 0.1
    s.axes_manager[3].scale = 0.1
    return s


def simulated_constant_shift_magnitude(beam_radius=10):
    """
    Create a simulated diffraction pattern with a direct beam that is shifted by a constant magnitude across
    the entire pattern. In the presence of electromagnetic fields in the entire sample area, the plane fitting
    can fail, however by minimizing the magnitude variance, we can fit a plane to the shifts


    Parameters
    ----------
    beam_radius : int, optional
        The radius of the direct beam in pixels. The default is 10.

    Returns
    -------
    probes : Diffraction2D
        A simulated diffraction pattern with a direct beam that is shifted by a constant
        magnitude across the entire pattern.

    """
    bright_field_disk = np.zeros((128, 128), dtype=np.int16)
    bright_field_disk[
        np.sum((np.mgrid[:128, :128] - 64) ** 2, axis=0) < beam_radius**2
    ] = 500

    probes = np.zeros((20, 20), dtype=int)
    probes = bright_field_disk[np.newaxis][probes]
    probes = pxm.signals.Diffraction2D(probes)

    p = [0.5] * 6  # Plane parameters
    x, y = np.meshgrid(np.arange(20), np.arange(20))
    base_plane_x = p[0] * x + p[1] * y + p[2]
    base_plane_y = p[3] * x + p[4] * y + p[5]

    base_plane = np.stack((base_plane_x, base_plane_y)).T
    data = base_plane.copy()

    shifts = np.zeros_like(data)
    shifts[:10, 10:] = (10, 10)
    shifts[:10, :10] = (10, -10)
    shifts[10:, 10:] = (-10, 10)
    shifts[10:, :10] = (-10, -10)
    data += shifts
    data = pxm.signals.BeamShift(data)
    probes.center_direct_beam(shifts=-data)
    probes.metadata["title"] = "Simulated Constant Shift Magnitude"
    return probes


def simulated_pn_junction(beam_shifts=None):
    """
    Create a simulated p-n junction diffraction pattern with alternating stripes.

    Parameters
    ----------
    shifts : array_like, optional
        The shifts to apply to the direct beam.  The default is None which
        corresponds to a shift of [[3, 3], [-3,3]] for the alternating stripes.
    Returns
    -------
    diffraction_pattern : Signal2D
        A simulated diffraction pattern with the direct beam shifted by .
    """
    if beam_shifts is None:
        beam_shifts = np.array([[4, 3]])
    data = np.zeros((30, 30, 256, 256))
    rr, cc = disk(center=(128, 128), radius=72)
    data[:, :, rr, cc] = 100
    data = np.random.poisson(data)  # Add Poisson noise
    s = pxm.signals.ElectronDiffraction2D(data)
    shifts = np.zeros((30, 30, 2))
    shifts[:, 12:18] = beam_shifts
    shifts = gaussian_filter(shifts, sigma=1.0)  # Smooth the shifts
    bs = pxm.signals.BeamShift(shifts)

    s.center_direct_beam(shifts=bs)  # Shift the direct beam
    s.axes_manager.signal_axes.set(
        scale=0.0025, units="nm^-1", offset=-0.0025 * 128, name=["kx", "ky"]
    )
    s.axes_manager.navigation_axes.set(
        name=["x", "y"], scale=[1, 1], units=["nm", "nm"]
    )
    s.metadata["title"] = "Simulated pn-junction"
    s.calibration.beam_energy = 200  # Set the beam energy to 200 keV
    return s
